# ----------------------------------------------------------------------
# Created: lör maj 15 17:21:15 2021 (+0200)
# Last-Updated:
# Filename: core.py
# Author: Yinan Yu
# Description:
# ----------------------------------------------------------------------

import uuid
import collections
from copy import deepcopy
import pydot
import matplotlib.image as mpimg
from io import BytesIO
from pprint import pprint
import json

from alex.alex import const, dsl_parser, util


# FIXME: use typing
def get_value_type(label):
    if label in const.INGREDIENT_TYPES \
       or label in const.INPUT_TYPES \
       or label in const.REGULARIZERS:
        node_type = "ingredient"
    elif label in const.ALL_RECIPES:
        node_type = "recipe"
    else:
        node_type = "hyperparam"
    return node_type


def get_node(tree, name):
    return tree[name]


def get_nodes(tree, names):
    return list(map(lambda x: tree[x] if x in tree else None, names))


def get_subtree(tree, name):
    tree = deepcopy(tree)
    subtree = collections.OrderedDict()
    descendants = tree[name]["descendants"]

    for node in descendants:
        subtree[node] = tree[node]
    subtree[name] = tree[name]
    return subtree


def get_children(root, tree, field=None):
    if field is not None:
        return list(map(lambda x: tree[x][field], tree[root]["children"]))
    else:
        return list(map(lambda x: tree[x], tree[root]["children"]))


def get_ancestor_param_node(node, tree, field=None):
    tree = deepcopy(tree)
    node = deepcopy(node)
    if node["value"] in const.ALL_PARAMS:
        parent_node = node
    elif node["name"] == "root" \
         or node["type"] == "ingredient" \
         or node["type"] == "recipe":
        parent_node = None
    else:
        parent = tree[node["parent"]]
        while parent["value"] not in const.ALL_PARAMS:
            _name = parent["name"]
            parent = tree[tree[_name]["parent"]]
            if parent["value"] == "root":
                parent = None
                break
        parent_node = parent
    if parent_node is None:
        return None
    elif field is None:
        return parent_node
    else:
        return parent_node[field]


def get_ancestor_ingredient_node(node, tree, field=None):
    name = node["meta"]["position"]["component"]
    if field is None:
        return tree[name]
    else:
        return tree[name][field]


def get_parent(root, tree, field=None):
    parent_node = tree[tree[root]["parent"]]
    if field is not None:
        parent_node = parent_node[field]
    return parent_node


def get_descendants(root, tree, field=None):
    if field is not None:
        return list(map(lambda x: tree[x][field], tree[root]["descendants"]))
    else:
        return list(map(lambda x: tree[x], tree[root]["descendants"]))


def get_labels(tree):
    return list(map(lambda x: tree[x]["label"], tree))


def name_to_index(names, tree):
    name_list = list(tree.keys())
    return list(map(lambda x: name_list.index(x), names))


def get_child_by_value(value, root, tree):
    for child in get_children(root, tree):
        if child["value"] == value:
            return child


def get_descendant_args_by_value(value, root, tree):
    for descendant in get_descendants(root, tree):
        if descendant["value"] == value:
            children = get_children(descendant["name"], tree, "value")
            return children


def instantiate_node(name, label, value, meta, parent, children, descendants):
    node = dict()
    node["name"] = name
    node["label"] = label
    node["value"] = value
    node["children"] = deepcopy(children)
    node["descendants"] = deepcopy(descendants)
    node["parent"] = parent
    node["lmd"] = descendants[0] if len(descendants)>=1 else name
    node["type"] = get_value_type(value)
    node["meta"] = meta
    return node


import ast
import traceback


def get_param(params, param):
    try:
        params = ast.literal_eval(params)
    except:
        # print("Error when parsing", params)
        # traceback.print_exc()
        # params is treated as a string literal
        pass
    if not isinstance(params, dict):
        if param == "name":
            return "%s_%s" % (str(params), str(uuid.uuid1()))
        else:
            return params
    if param not in params:
        raise Exception("param %s not recgonized" % str(param))
    return params[param]


def alex_graph_to_json(graph,
                       root_name="root",
                       json_obj=collections.OrderedDict(),
                       naive=True,
                       label_path=None,
                       position: dict = {"inputs": None,
                                         "component": None,
                                         "shape": None,
                                         "input_shape": None}):
    """
    This function converts alex graph into a json object.
    A node can be a recipe, an ingredient or a hyperparameter
    All info of a node is encoded into a json sstring

    """
    if isinstance(graph, dict) and ("visible" in graph) and not graph["visible"]:
        return graph, json_obj

    graph = deepcopy(graph)
    json_obj = collections.OrderedDict(sorted(json_obj.items()))
    if isinstance(graph, dict) \
       and "subgraph" in graph \
       and isinstance(graph["subgraph"], dict): # is a recipe
        _position = {"inputs": graph["inputs"],
                     "component": root_name,
                     "shape": None,
                     "input_shape": None}
        label = deepcopy(graph["type"])
        root_name = {"value": graph["type"],
                     "label": label,
                     "name": root_name,
                     "meta": {"hyperparams": {},
                              "position": _position,
                              "label_path": label_path}}
        root_name = str(root_name)
        json_obj[root_name] = collections.OrderedDict()
        for _name, _graph in graph["subgraph"].items():
            _position = {"inputs": _graph["inputs"],
                         "component": _name,
                         "shape": None,
                         "input_shape": None}
            _json_obj = alex_graph_to_json(_graph,
                                           _name,
                                           collections.OrderedDict(),
                                           naive=naive,
                                           label_path=None,
                                           position=_position)
            json_obj[root_name].update(_json_obj)
    else: # if is an ingredient or hyperparameters
        if isinstance(graph, dict) and "hyperparams" in graph:
            if len(graph["hyperparams"]) != 0: # if is an ingredient with hyperparameter

                if not naive:
                    hyperparam_str = str(uuid.uuid3(uuid.NAMESPACE_DNS,
                                                    json.dumps(graph["hyperparams"],
                                                               sort_keys=True)))
                    label = "%s#%s#" % (graph["type"], hyperparam_str)

                else:
                    label = graph["type"]
                _position = {"inputs": graph["inputs"],
                             "component": root_name,
                             "shape": graph["meta"]["shape"],
                             "input_shape": graph["meta"]["input_shape"]}
                _root_name = root_name
                root_name = {"value": graph["type"],
                             "label": label,
                             "name": root_name,
                             "meta": {"hyperparams": graph["hyperparams"],
                                      "position": _position,
                                      "label_path": label_path,
                                      "block": graph["meta"]["block"]}}

                root_name = str(root_name)
                json_obj[root_name] = collections.OrderedDict()
                for _name in sorted(graph["hyperparams"]):
                    _graph = graph["hyperparams"][_name]
                    _name = "%s/%s" % (_root_name, _name)
                    if "#" in label:
                        label_path = label.split("#")[0] + label.split("#")[2]
                    else:
                        label_path = label
                    _json_obj = alex_graph_to_json(_graph,
                                                   _name,
                                                   collections.OrderedDict(),
                                                   naive=naive,
                                                   label_path=label_path,
                                                   position=_position)
                    json_obj[root_name].update(_json_obj)

            else:
                _position = {"inputs": graph["inputs"],
                             "component": root_name,
                             "shape": graph["meta"]["shape"],
                             "input_shape": graph["meta"]["input_shape"]}

                root_name = {"value": graph["type"],
                             "label": graph["type"],
                             "name": root_name,
                             "meta": {"hyperparams": {},
                                      "position": _position,
                                      "label_path": label_path,
                                      "block": graph["meta"]["block"]}}
                root_name = str(root_name)

                json_obj[root_name] = None

        else: # hyperparam tree
            _root_name = deepcopy(root_name)
            _value = _root_name.split("/")[-1]
            label_path = "%s/%s" % (label_path, _value)
            root_name = {"value": _value,
                         "label": label_path,
                         "name":  _root_name,
                         "meta": {"hyperparams": graph,
                                  "position": position,
                                  "label_path": label_path}}
            root_name = str(root_name)
            if isinstance(graph, dict):

                json_obj[root_name] = collections.OrderedDict()
                _graph = sorted(graph)
                for _name in _graph:
                    __name = "%s/%s" % (_root_name, _name)
                    _json_obj = alex_graph_to_json(graph[_name],
                                                   __name,
                                                   collections.OrderedDict(),
                                                   naive=naive,
                                                   label_path=label_path,
                                                   position=position)
                    json_obj[root_name].update(_json_obj)

            elif isinstance(graph, list):
                json_obj[root_name] = []
                for i, _graph in enumerate(graph):
                    _label = _graph
                    _name = "%s/%s_%i" % (_root_name,
                                          _label, i)
                    label_path = "%s/%s" % (label_path, _label)

                    _json_value = {"value": _label,
                                   "label": label_path,
                                   "name": _name,
                                   "meta": {"hyperparams": graph,
                                            "position": position,
                                            "label_path": label_path}}
                    json_obj[root_name].append(str(_json_value))
            else:
                _label = graph
                _name = "%s/%s" % (_root_name,
                                   _label)
                label_path = "%s/%s" % (label_path, _label)
                _json_value = {"value": _label,
                               "label": label_path,
                               "name": _name,
                               "meta": {"hyperparams": graph,
                                        "position": position,
                                        "label_path": label_path}}

                json_obj[root_name] = str(_json_value)


    return json_obj


def json_to_tree(json,
                 root=None,
                 name_root="root",
                 parent_name=None,
                 tree=collections.OrderedDict(),
                 exclude_types=[]):
    json = deepcopy(json)
    tree = deepcopy(tree)

    children = []
    descendants = []
    if root is not None and get_value_type(get_param(root, "value")) in exclude_types:
        return tree

    if type(json) == dict:
        json = collections.OrderedDict(sorted(json.items()))
    for node in json:
        _json = json[node]
        name = get_param(node, "name")
        if get_value_type(get_param(node, "value")) in exclude_types:
            continue
        _descendants = []
        if isinstance(_json, dict):
            tree = json_to_tree(_json,
                                node,
                                name,
                                name_root,
                                tree,
                                exclude_types=exclude_types)
            _descendants = tree[name]["descendants"]
        else:
            if isinstance(_json, list):
                for _item in _json:
                    _name = get_param(str(_item), "name")
                    _label = get_param(str(_item), "label")
                    _meta = get_param(str(_item), "meta")
                    _value = get_param(str(_item), "value")
                    if get_value_type(_label) not in exclude_types:
                        tree[_name] = instantiate_node(_name,
                                                       _label,
                                                       _value,
                                                       _meta,
                                                       name,
                                                       [], [])
                        _descendants.append(_name)
            else:
                if _json is not None:
                    _name = get_param(str(_json), "name")
                    _label = get_param(str(_json), "label")
                    _meta = get_param(str(_json), "meta")
                    _value = get_param(str(_json), "value")
                    if get_value_type(_label) not in exclude_types:
                        tree[_name] = instantiate_node(_name,
                                                       _label,
                                                       _value,
                                                       _meta,
                                                       name,
                                                       [], [])
                        _descendants = [_name]

            tree[name] = instantiate_node(name,
                                          get_param(node, "label"),
                                          get_param(node, "value"),
                                          get_param(node, "meta"),
                                          name_root,
                                          _descendants,
                                          _descendants)
        if name in tree:
            descendants += _descendants + [name]
            children.append(name)
    if root is not None:
        tree[name_root] = instantiate_node(name_root,
                                           get_param(root, "label"),
                                           get_param(root, "value"),
                                           get_param(root, "meta"),
                                           parent_name,
                                           children,
                                           descendants)
    return tree


def alex_graph_to_tree(network_graph, exclude_types=[], naive=True):
    json_obj = alex_graph_to_json(network_graph, naive=naive)

    tree = json_to_tree(json_obj,
                        exclude_types=exclude_types)
    return tree


def draw(tree, graph_path='example.png', annotation=dict(), dpi=800, size=5, label_field="value", excluded_types=[]):
    graph = pydot.Dot(graph_type='digraph', rankdir='LR', dpi=dpi, size=size, fontsize=12)
    for component in tree:
        name = component
        if tree[component]["type"] in excluded_types:
            continue
        if name not in graph.obj_dict['nodes'].keys():
            if name in annotation:
                color = annotation[name]
            else:
                color = "white"
            if label_field not in tree[component]:
                label = str(tree[component]["label"])
            else:
                label = str(tree[component][label_field])
            node = pydot.Node(name,
                              label=label,
                              shape="box",
                              style='filled',
                              fillcolor=color)
            graph.add_node(node)
        children_list = tree[component]["children"].copy()
        for child in children_list:
            if tree[child]["type"] in excluded_types:
                continue

            if child not in graph.obj_dict['nodes'].keys():
                if child in annotation:
                    color = annotation[child]
                else:
                    color = "white"

                if label_field not in tree[child]:
                    label = str(tree[child]["label"])
                else:
                    label = str(tree[child][label_field])
                node = pydot.Node(child,
                                  label=label,
                                  shape="box",
                                  style='filled',
                                  fillcolor=color)
                # if isinstance(tree[child]["meta"], dict):
                #     if tree[child]["name"] != "root" and "block" in tree[child]["meta"]:
                #         block = tree[child]["meta"]["block"]
                #         blocks[block].add_node(node)
                #     else:
                          # graph.add_node(node)
                graph.add_node(node)

            edge = pydot.Edge(component, child)
            graph.add_edge(edge)

    if graph_path is None:
        png = graph.create_png(prog="dot")
        bio = BytesIO()
        bio.write(png)
        bio.seek(0)
        graph = mpimg.imread(bio)
    else:
        graph.write_png(graph_path)
    return graph



# Created: sön nov  12 10:11:23 2017 (+0100)
# ----------------------------------------------------------------------
# Last-Updated:
# Filename: parser.py
# Author: Yinan
# Description: This module parses block of components (recipes) and basic components (ingredients) from .yml files.
# ----------------------------------------------------------------------


import pydot
from copy import deepcopy
import os
import string
import collections.abc as collections
from collections import OrderedDict, Iterable
from pprint import pprint
import traceback

from typing import Union

from alex.alex import checkpoint, const, util, validation


strs = list(string.ascii_lowercase)

# FIXME: find a better solution
# These are used to set up names
global global_count0, global_count1, component_count, prev_component
global_count0 = 0
global_count1 = 0
component_count = 0

prev_component = None

# Each keywords in the yml file is a function
"""
repeat: component -> [component]
share: component_name -> source_name -> [var_names]
"""

"""
The parsing process involves several passes in order to generate the final ast
At each pass, an intermediate data structure is generated
"""
Fns1pass = (const.REPEAT, const.NAME, const.INPUTS)
Fns2pass = (const.PARAMS_TRAINING, "share")
Fns = Fns2pass + Fns1pass
Training = (const.TYPE)
Values = (const.HYPERPARAMS, const.VAR, const.STATS, const.TENSOR) # output
Meta = (const.NAME, const.INPUTS, const.TYPE, const.DTYPE, const.SHAPE)


######## parse
def alex_reader(user_defined):

    def _read_inputs(component):
        component = clone(component)
        if const.INPUTS not in component:
            # If no inputs is specified, inputs is set to the previous component (sequential)
            component = {**{const.INPUTS: None}, **component}
        elif not isinstance(component[const.INPUTS], list):
            # Convert all inputs into list
            component[const.INPUTS] = [component[const.INPUTS]]
        return component

    def _read_hyperparams(component):
        component = clone(component)
        hyperparams = dict() if const.HYPERPARAMS not in component else clone(component[const.HYPERPARAMS])
        component_type = component[const.TYPE]
        if component_type in const.PARAMS:
            for param in const.PARAMS[component_type]:
                if param not in hyperparams:
                    hyperparams[param] = dict()
                if const.DTYPE not in hyperparams[param]:
                    hyperparams[param][const.DTYPE] = const.DEFAULT_DTYPE
        component[const.HYPERPARAMS] = hyperparams
        return component

    if "default" not in user_defined:
        default = {"dtype": const.DEFAULT_DTYPE}
    else:
        default = user_defined["default"]

    # Set up default value
    reader = {const.NAME: lambda x: {**{const.NAME: None}, **x} if const.NAME not in x else clone(x),
              const.INPUTS: _read_inputs,
              const.HYPERPARAMS: _read_hyperparams,
              "visible": lambda x: {**{"visible": True}, **x} if "visible" not in x else clone(x),
              "dtype": lambda x: {**{"dtype": default["dtype"]}, **x} if "dtype" not in x else clone(x)}
    return reader


def fill_mandatory(component, reader):
    for fn in reader:
        component = reader[fn](component)
    return component


def clone(component):
    return deepcopy(component)


def recipe_defined(rtype):
    # TODO: temporary. Use types instead
    return rtype not in list(const.DET_COMPONENTS_NO_HYPE.keys()) + [const.DATA,
                                                                 const.LABEL]

# FIXME
def _resolve_hyperparams(hyperparams):
    return hyperparams


def _parse_ingredient_hyperparams(hyperparams):
    hyperparams = clone(hyperparams)
    for hyperparam in hyperparams:
        if isinstance(hyperparams[hyperparam], dict):
            hyperparam_file = os.path.join(const.COMPONENT_BASE_PATH,
                                           hyperparam+".yml")
            if os.path.exists(hyperparam_file):
                _hyperparam = util.read_yaml(hyperparam_file)
                util.fill_dict(_hyperparam, hyperparams[hyperparam])
            _hyperparam = _parse_ingredient_hyperparams(hyperparams[hyperparam])
            hyperparams[hyperparam] = clone(_hyperparam)
    return hyperparams


def _config_to_inter_graph(user_defined, reader, config=None, recipe=None):
    if config is None:
        config = user_defined
    if recipe is None:
        recipe = const.ROOT
    # If configuration is not an ingredient
    if const.PARAMS_NETWORK in config:
        component = dict()
        if const.PARAMS_NETWORK not in user_defined:
            component = fill_mandatory(clone(user_defined), reader)
        component[const.HYPERPARAMS] = []
        component[const.TYPE] = recipe
        for subconfig in config[const.PARAMS_NETWORK]:
            component[const.HYPERPARAMS].append(dict())
            component[const.HYPERPARAMS][-1] = deepcopy(subconfig)
            if recipe_defined(subconfig[const.TYPE]):
                subconfig_def = util.read_yaml(os.path.join(const.COMPONENT_BASE_PATH,
                                                            subconfig[const.TYPE] + const.YAML))
            else:
                subconfig_def = {}
            component[const.HYPERPARAMS][-1] = _config_to_inter_graph(subconfig,
                                                                      reader,
                                                                      subconfig_def,
                                                                      subconfig[const.TYPE])
    else:
        # TODO: make it recursive
        component = fill_mandatory(clone(user_defined), reader)
        hyperparams = clone(config)
        util.fill_dict(hyperparams, component[const.HYPERPARAMS])
        component[const.HYPERPARAMS] = _parse_ingredient_hyperparams(component[const.HYPERPARAMS])
    return component


def _inter_graph_valid(inter_graph):
    # TODO: refine
    # check validity
    return True


def config_to_inter_graph(user_defined, reader):
    graph = _config_to_inter_graph(user_defined, reader, config=None, recipe=None)
    _inter_graph_valid(graph)
    return graph


###############################################################################
def _get_node(component):
    graph = dict()
    graph[const.VALUE] = dict()
    graph[const.META] = dict()
    for v in Values:
        if v in component:
            graph[const.VALUE][v] = clone(component[v])
        else:
            graph[const.VALUE][v] = dict()
    for m in Meta:
        if m in component:
            graph[const.META][m] = component[m]
    graph[const.META][const.NAME] = None
    graph[const.META]["visible"] = component["visible"]
    graph[const.META]["dtype"] = component["dtype"]
    graph[const.META][const.SCOPE] = component[const.SCOPE] if const.SCOPE in component else component[const.TYPE]
    graph[const.VALUE][const.VALUE] = dict()
    # if const.PARAMS_TRAINING in component: # not used for now... maybe for variables
    #     graph[const.VALUE][const.HYPERPARAMS][const.PARAMS_TRAINING] = component[const.PARAMS_TRAINING]
    return graph


def _parse_to_ast(component, args):
    ast = clone(args)
    for fn in Fns1pass:
        if fn in component:
            ast = (fn, component[fn], ast)
    return ast


def inter_graph_to_ast(component):
    component = clone(component)
    if isinstance(component[const.HYPERPARAMS], list): # not atomic
        args = []
        for _component in component[const.HYPERPARAMS]:
            _component[const.SCOPE] = component[const.TYPE]
            args.append(inter_graph_to_ast(_component))
    else:
        args = [_get_node(component)]
    ast = _parse_to_ast(component, args)
    return ast


######## eval
############################ Begin the first pass #######################
def _append(lst, elm):
    if isinstance(elm, list):
        return lst + elm
    else:
        return lst + [elm]


def _get_name(name, append=False):
    global global_count0, global_count1, component_count
    _name = str(component_count) \
        + strs[global_count0 % len(strs)] \
        + strs[global_count1 % len(strs)]
    if name:
        if append:
            _name = "%s_%s"%(name, _name)
        else:
            _name = name
    global_count0 += 1
    global_count1 += 8
    component_count += 1
    return _name


def _mark_repeat_idx(components, idx, times, count):
    components = clone(components)
    for i, component in enumerate(components):
        component[const.META][const.RPT_IDX] = {"block_repeat_idx": idx,
                                                "max_block_repeat": times-1,
                                                "component_idx_in_block": i,
                                                "component_count": count,
                                                "max_components": times*len(components)}
    return components


def eval_repeat(times, components):
    _components = clone(components)
    components = []
    count = 0
    for idx in range(times):
        count += 1
        __components = clone(_components)
        __components = _mark_repeat_idx(__components, idx, times, count)
        components = _append(components, __components)
    return components


def eval_name(name, components):
    components = clone(components)
    top_name = _get_name(name)
    if components[0][const.META][const.SCOPE] == const.ROOT:
        scope = ""
    else:
        scope = "%s_" % components[0][const.META][const.SCOPE]

    for component in components:
        is_atomic = False
        __name = clone(component[const.META][const.NAME])  # name resolved so far; might add prefix (path) to it
        _name = top_name  # suffix, e.g. "xyz123" in conv_xyz123
        if __name is None:  # it is the lowest level - component name has not been resolved yet
            is_atomic = True
            __name = _get_name(name)

        if const.RPT_IDX in component[const.META]:  # repeat
            rpt_info = component[const.META][const.RPT_IDX]
            rpt_idx = rpt_info["block_repeat_idx"]
            max_rpt = rpt_info["max_block_repeat"]
            if rpt_idx != max_rpt:
                if is_atomic:
                    __name = "%s_%i" % (__name, rpt_idx)
                else:
                    _name = "%s_%i" % (top_name, rpt_idx)
        if is_atomic:
            if name is None:  # if no user defined name
                __name = "%s_%s"%(component[const.META][const.TYPE], __name)  # const.TYPE + "_" + "id_string"
        else: # when component is not on the lowest level
            if name is None:
                __name = "%s%s/%s" % (scope, _name, __name)
            else: # user defined name in recipe
                __name = "%s/%s" % (_name, __name)
        component[const.META][const.NAME] = __name
    return components


def eval_inputs(inputs, components):
    global prev_component

    components = clone(components)
    user_defined = True
    if inputs is None:
        user_defined = False
        inputs = prev_component
    _prev_component = inputs

    for i, component in enumerate(components):
        component = clone(component)
        if component[const.META][const.TYPE] == const.DATA or component[const.META][const.TYPE] == "label":
            continue
        if i == 0 and user_defined:
            component[const.META][const.INPUTS] = inputs
        else:
            # If input components are not defined, then make it sequential
            if component[const.META][const.INPUTS] is None:
                component[const.META][const.INPUTS] = [_prev_component]
            else:
                if i != 0:
                    prev_dir = _prev_component.replace(_prev_component.split("/")[-1], "")
                    for _b, _base in enumerate(component[const.META][const.INPUTS]):
                        if isinstance(_base, dict):
                            pass
                        base_name = _base.split("/")[-1]
                        _in = prev_dir + base_name
                        component[const.META][const.INPUTS][_b] = _in

        if const.RPT_IDX in component[const.META]:
            rpt_info = component[const.META][const.RPT_IDX]
            _component_count = rpt_info["component_count"]
            component_idx_in_block = rpt_info["component_idx_in_block"]
            if _component_count != 1 and component_idx_in_block == 0:
                if isinstance(_prev_component, list):
                    component[const.META][const.INPUTS] = _prev_component
                else:
                    component[const.META][const.INPUTS] = [_prev_component]
            component[const.META].pop(const.RPT_IDX, None)
        _prev_component = component[const.META][const.NAME]
        components[i] = component

    prev_component = _prev_component
    return components


############################ The end of the first pass #######################
############################ Begin second pass ###############################
####### Second pass only deal with local dependencies

### TODO: share
def eval_share(share, components):

    # for component in components:
    #     if share is not None:

    #         if component[""][]SHARABLE
    return components


############################ The end of the second pass #######################
def std_env():
    env = {const.REPEAT: eval_repeat,
           const.NAME: eval_name,
           const.INPUTS: eval_inputs,
           "share": lambda t, x: deepcopy(x), # TODO: not implemented
    }
    return env


env = std_env()


def eval_ast_to_graph(ast, env=env):
    ast = clone(ast)
    if isinstance(ast, list):
        graph = []
        for _ast in ast:
            graph = _append(graph, eval_ast_to_graph(_ast, env=env))
    if isinstance(ast, tuple):
        if isinstance(ast[2], list):
            if all(isinstance(x, dict) for x in ast[2]):
                graph = env[ast[0]](ast[1], ast[2])
            else:
                graph = []
                for _ast in ast[2]:
                    _graph = eval_ast_to_graph(_ast, env=env)
                    graph = _append(graph, _graph)
                graph = eval_ast_to_graph((ast[0], ast[1], graph), env=env)
        elif isinstance(ast[2], tuple):
            graph = eval_ast_to_graph(ast[2], env=env)
            graph = eval_ast_to_graph((ast[0], ast[1], graph), env=env)
    return graph

# TODO: merge into the same module as schedulers
# Functions for inputs
def _parse_range(range):
    _range = range.split(":")
    lower = _range[0]
    upper = _range[1]
    if lower != "":
        lower = int(lower)
    if upper != "":
        upper = int(upper)
    return lower, upper


def find_key(data, key):
    if key in data: return data[key]
    found = None
    for _key, value in data.items():
        if _key == "connect_to":
            continue
        if isinstance(value, dict):
            found = find_key(value, key)
        if found is not None:
            break
    return found


def delete_key(data, key):
    if key in data:
        del data[key]

    for _key, value in data.items():
        if isinstance(value, dict):
            delete_key(value, key)


def connect_to(components, current_component_type, current_component_hyperparams=None, range=None, types=None, scope="all"):

    def _get_name_scope(_component):
        return "/".join(_component["meta"]["name"].split("/")[:-1])

    def _get_type(_component):
        return _component["meta"]["type"]

    if scope != "all" and (scope != "current"):
        raise Exception("scope has to be 'current' or 'all'!")
    if current_component_hyperparams is None:
        current_component_hyperparams = components[-1]["value"]["hyperparams"]
    current_component_name_scope = _get_name_scope(components[-1])
    if range is not None:
        lower, upper = _parse_range(range)
    else:
        lower, upper = -2, -1

    connected_component = []
    if lower == "" and upper != "":
        _components = components[:upper]
    elif upper == "" and lower != "":
        _components = components[lower:]
    elif upper != "" and lower != "":
        _components = components[lower:upper]
    else:
        _components = components

    for component in _components:
        type_match = True
        scope_match = True
        _component_type = _get_type(component)
        if types is not None and (_component_type not in types):
            type_match = False
        if scope == "current" and (current_component_name_scope != (_get_name_scope(component))):
            scope_match = False
        if type_match and scope_match:
            connected_component.append("%s/%s" % (component["meta"]["name"],
                                                  types[_component_type]))
            # Make regularization part of the hyperparam subtree
            # component["value"]["hyperparams"][types[_component_type]][current_component_type] = current_component_hyperparams
        delete_key(component["value"]["hyperparams"], "connect_to")
    delete_key(current_component_hyperparams, "connect_to")
    return connected_component


# TODO: implement
def _feature_maps_(range, type):
    pass

def flatten(lst):
    flattened = []
    for obj in lst:
        if not isinstance(obj, list):
            obj = [obj]
        flattened += obj
    return flattened


def global_update(components):
    """
    Here we resolve any functions defined for inputs
    """
    prev_dir = None
    prev_level = 0
    for icomponent, component in enumerate(components):
        _dir = component[const.META][const.NAME].split("/")
        current_level = len(_dir) - 1
        current_dir = component[const.META][const.NAME].replace(_dir[-1], "")
        if current_dir != prev_dir and prev_dir is not None:
            if prev_level <= current_level:
                inputs = component[const.META][const.INPUTS][0]
        if component[const.META][const.INPUTS] is not None:
            need_flatten = False
            for i, _input in enumerate(component[const.META][const.INPUTS]):
                if _input is None:
                    component[const.META][const.INPUTS][i] = const.INPUTS
                elif const.INPUTS in _input:
                    try:
                        component[const.META][const.INPUTS][i] = inputs
                    except: # if inputs is not defined
                        component[const.META][const.INPUTS][i] = const.INPUTS

                elif isinstance(_input, dict):
                    if "connect_to" in _input:
                        component[const.META][const.INPUTS][i] = connect_to(components[:icomponent+1],
                                                                            component[const.META][const.TYPE],
                                                                            **_input["connect_to"])
                    need_flatten = True
            if need_flatten:
                component[const.META][const.INPUTS] = flatten(component[const.META][const.INPUTS])
        component[const.VALUE][const.HYPERPARAMS] = _resolve_hyperparams(component[const.VALUE][const.HYPERPARAMS])
        for _key in component[const.VALUE][const.HYPERPARAMS]:
            hyperparam = component[const.VALUE][const.HYPERPARAMS][_key]
            # TODO: make this nested?
            if isinstance(hyperparam, dict) and ("connect_to" in hyperparam):
                connect_to(components[:icomponent+1],
                           component[const.META][const.TYPE],
                           hyperparam,
                           **hyperparam["connect_to"])
        prev_level = current_level
        prev_dir = current_dir

    return components



########### Params_Training ###############
def _user_fn(fn, kwargs):
    return lambda x: const.USER_FNS[fn](x, **kwargs)


######################################### Different views:
def draw_graph(graph_list, level=2, graph_path='example.png', show="name"):
    graph = pydot.Dot(graph_type='digraph', rankdir='LR', dpi=800, size=5, fontsize=18)

    added = []
    for component in graph_list:
        _dir = component[const.META][const.NAME].split("/")
        _dir = _dir[:level]
        name = "/".join(_dir)
        if name in added:
            continue
        added.append(name)
        node = pydot.Node(name, shape="box")
        if show == "name":
            label = name.split("/")[-1]# .split("_")[0]
        else:
            label = component["meta"]["type"]
        node.set_label(label)
        graph.add_node(node)

        if not component[const.META][const.TYPE] in const.INPUT_TYPES:
            inputs = component[const.META][const.INPUTS]
            for _input in inputs:
                __input = "/".join(_input.split("/")[:level])
                edge = pydot.Edge(_input, name)
                graph.add_edge(edge)

    graph.write_png(graph_path)
    return graph


def _get_dir_name(path):
    _dir = path.split("/")
    scope = path.replace(_dir[-1], "")
    return scope, _dir[-1]


def list_to_graph(components, parent_level=1, parent_scope=""):
    network = dict()
    _subgraph = OrderedDict()
    network[const.LEVEL] = parent_level-1
    network[const.META] = dict()
    # TODO: remove the following info (cached in meta)
    network[const.INPUTS] = None
    network["visible"] = True
    network[const.TYPE] = const.ROOT
    i = 0
    while i < len(components):
        component = components[i]
        hyperparams = component[const.VALUE][const.HYPERPARAMS]
        _dir = component[const.META][const.NAME].split("/")
        level = len(_dir)
        scope, _name = _get_dir_name(component[const.META][const.NAME])
        if scope == "":
            _scope = _name
        else:
            _scope = scope + component[const.META][const.NAME].replace(scope, "").split("/")[0]
        inputs = component[const.META][const.INPUTS]
        if parent_scope!=scope:
            if level>parent_level:
                _components = []
                for _component in components[i:]:
                    __scope, __ = _get_dir_name(_component[const.META][const.NAME])
                    if scope in __scope:
                        _components.append(_component)
                        i += 1
                    else:
                        break
                _network = list_to_graph(_components, level, scope)
                _network[const.INPUTS] = inputs
                _network[const.HYPERPARAMS] = dict() # hyperparams for recipes
                _network[const.TYPE] = component[const.META][const.SCOPE]
                _network["visible"] = component[const.META]["visible"]
                _network[const.META] = component[const.META]
                # if _network[const.META]["type"] in const.COMPONENT_RECIPES:
                #     _network[const.META]["block"] = "model"

                _subgraph[scope[:-1]] = _network
            else:
                break
        else:
            _network = dict()
            _network[const.SUBGRAPH] = component[const.META][const.NAME]
            _network[const.TYPE] = component[const.META][const.TYPE]
            _network["visible"] = component[const.META]["visible"]
            _network[const.META] = component[const.META]
            _network["dtype"] = component[const.META]["dtype"]
            _network[const.INPUTS] = inputs
            _network[const.LEVEL] = level
            _network[const.HYPERPARAMS] = hyperparams

            _subgraph[_scope] = _network
            i += 1

    network[const.SUBGRAPH] = _subgraph
    return network


def draw_hyperparam_tree(graph,
                         graph_path="hypertree.png"):

    def _is_component(node):
        return isinstance(node, dict) and (const.SUBGRAPH in node)

    def _is_recipe(node):
        return isinstance(node[const.SUBGRAPH], OrderedDict)

    def _is_simple(graph):
        return not isinstance(graph, Iterable) or (isinstance(graph, str))

    def _add_node(parent_node, label, name, graph):
        if label in const.COMPONENTS:
            color = colors["COMPONENTS"]
        elif label in const.DET_COMPONENTS_HYPE:
            color = colors["DET_COMPONENTS_HYPE"]
        elif label in const.DET_COMPONENTS_NO_HYPE:
            color = colors["DET_COMPONENTS_NO_HYPE"]
        elif label in const.INFERENCE:
            color = colors["INFERENCE"]
        elif label in const.ALL_TRAINABLE_PARAMS:
            color = colors["TRAINABLE_PARAMS"]
        elif label in const.ALL_OTHER_PARAMS:
            color = colors["NONTRAINABLE_PARAMS"]
        else:
            color = "white"
        node = pydot.Node(name,
                          shape="box",
                          style="filled",
                          fillcolor=color,
                          fontsize=16)
        node.set_label(label)
        graph.add_node(node)
        edge = pydot.Edge(parent_node, name)
        graph.add_edge(edge)
        return graph

    def _draw_hyperparam_tree(_graph,
                              parent_node: str,
                              graph):

        _graph = clone(_graph)
        if _is_simple(_graph):
            graph = _add_node(parent_node=parent_node,
                              label=str(_graph),
                              name="%s/%s" % (parent_node, str(_graph)),
                              graph=graph)
        else:
            for i, _node in enumerate(_graph):
                """ The assumption here is that _graph can be either a dict
                        or hyperparams and hyperparams have a very simple
                        structure, which means that _node is either a string
                        or a scalar. If _graph is a list and _node is a complex
                        data structure, it is not handled.
                """

                if _is_simple(_node):
                    label = str(_node) # node is simple
                else:
                    print(_node)
                    raise Exception("Unrecognized node type")

                name = "%s/%s" % (parent_node, label)
                if isinstance(_graph, list):
                    name = "%s_%i" % (name, i)

                if isinstance(_graph, dict) and _is_component(_graph[_node]):
                    # FIXME:
                    if not _graph[_node]["visible"]:
                        continue
                    label = clone(_graph[_node][const.TYPE])

                graph = _add_node(parent_node, label, name, graph)

                if isinstance(_graph, dict):
                    _node = clone(_graph[_node])
                    if _is_simple(_node):
                        graph = _add_node(parent_node=name,
                                          label=str(_node),
                                          name="%s/%s" % (name, str(_node)),
                                          graph=graph)
                    elif isinstance(_node, list):
                        for ii, __node in enumerate(_node):
                            graph = _add_node(parent_node=name,
                                              label=str(__node),
                                              name="%s/%s_%i" % (name, str(__node), ii),
                                              graph=graph)


                if isinstance(_node, dict):
                    # component or hyperparams
                    if _is_component(_node): # is a component (ingredient or recipe)
                        if _is_recipe(_node):
                            graph = _draw_hyperparam_tree(_node[const.SUBGRAPH],
                                                          parent_node=name,
                                                          graph=graph)
                        # go in the hyperparam tree
                        if recipe_defined(_node[const.TYPE]):
                            graph = _draw_hyperparam_tree(_node[const.HYPERPARAMS],
                                                          parent_node=name,
                                                          graph=graph)
                    else: # in the hyperparam tree
                        graph = _draw_hyperparam_tree(_node,
                                                      parent_node=name,
                                                      graph=graph)

        return graph

    colors = {"COMPONENTS": "red",
              "DET_COMPONENTS_HYPE": "green",
              "DET_COMPONENTS_NO_HYPE": "yellow",
              "INFERENCE": "cyan",
              "NONTRAINABLE_PARAMS": "#8ecae6",
              "TRAINABLE_PARAMS": "#ffafcc"}
    _graph = list_to_graph(graph)
    dot_graph = pydot.Dot(graph_type='digraph', rankdir='LR', dpi=100, size=100, fontsize=18)
    parent_node = const.ROOT
    dot_graph = _draw_hyperparam_tree(_graph[const.SUBGRAPH], parent_node, dot_graph)
    dot_graph.write(graph_path, format="png")


def list_to_dict(components):
    components = deepcopy(components)
    components = OrderedDict(map(lambda x:
                                 (components[x]["meta"]["name"],
                                  components[x]), range(len(components))))

    return components


### d3
### FIXME: test this
def graph_to_d3_ast(graph):
    graph = clone(graph)
    children = []
    for node_name in graph:
        _node = graph[node_name]
        node = dict()
        node["name"] = node_name

        if isinstance(_node, dict) and "level" in _node:
            node["level"] = _node["level"]
            node["type"] = _node["type"]
            if isinstance(_node["subgraph"], dict):
                node["children"] = graph_to_d3_ast(_node["subgraph"])
            if len(_node["hyperparams"]) != 0:
                node["children"] = graph_to_d3_ast(_node["hyperparams"])
        elif isinstance(_node, dict): # is part of hyperparams
            node["type"] = "hyperparams"
            node["children"] = graph_to_d3_ast(_node)
        elif isinstance(_node, list):
            if "type" not in node:
                node["type"] = "hyperparams"
            node["children"] = []
            for __node in _node:
                node["children"].append({"name": __node, "type": "hyperparams"})
        elif _node is None:
            if "type" not in node:
                node["type"] = "hyperparams"
            node["children"] = [{"name": "none", "type": "hyperparams"}]
        else:
            node["children"] = [{"name": _node, "type": "hyperparams"}]
            if "type" not in node:
                node["type"] = "hyperparams"
        children.append(node)
    return children


def annotate(components, lazy=True):
    schemas = dict()
    components = list_to_dict(components)
    for name in components:
        component = components[name]
        component_type = component["meta"]["type"]
        validator = validation.get(component)
        try:
            if component_type not in schemas:
                schemas[component_type] = validator.load_schema(component_type)
            if schemas[component_type] is not None:
                validator.validate_schema(component, schemas[component_type])

            component["meta"]["input_shape"] = validator.get_input_shape(component, components)
            component["meta"]["shape"] = validator.get_shape(component, components)
            if not lazy:
                validator.eval_input_channels(component, components)
                validator.validate_connection(component, components)


            inputs = component["meta"]["inputs"]
            # Annotate block: data, model, loss, optimizer
            # TODO: move this to nodes
            if inputs is not None and "inputs" in inputs:
                component["meta"]["block"] = "model"
            elif component["meta"]["type"] in const.INPUT_TYPES:
                component["meta"]["block"] = "data"
            elif component["meta"]["type"] not in const.OPTIMIZER_BLOCK \
                 and (component["meta"]["type"] in const.LOSS_BLOCK \
                      or len(list(filter(lambda x: components[x]["meta"]["block"] == "loss",
                                         inputs)))!=0):
                component["meta"]["block"] = "loss"
            elif component["meta"]["type"] in const.OPTIMIZER_BLOCK \
                 or len(list(filter(lambda x: components[x]["meta"]["block"] == "optimizer",
                                    inputs)))!=0:
                component["meta"]["block"] = "optimizer"
            else:
                component["meta"]["block"] = "model"
        except Exception:
            print("------------- Component %s annotation failed ----------" % (name))
            traceback.print_exc()
            raise
        components[name] = component
    return list(components.values())


def config_to_type(yml_file):
    return yml_file.split("/")[-1].split(".")[0]


################################################################################
# Public API
################################################################################
def parse(yml_file, return_dict=False, lazy=True):
    global global_count0, global_count1, component_count, prev_component
    global_count0 = 0
    global_count1 = 0
    component_count = 0
    # print("parsing, yml_file", yml_file)
    prev_component = None
    try:
        # Read the yml file
        user_defined = util.read_yaml(yml_file)
        reader = alex_reader(user_defined)
        # Step 1: transform yml into a (recursive) dict
        # Each node in the dict has the following keys:
        # hyperparams: list or dict, default: {}
        # type: string
        # inputs: list, default: None
        # name: string, default: None
        # (optional) dtype: data type
        # (optional) shape: list
        _graph:dict = config_to_inter_graph(user_defined, reader)
        # Step 2: transform _graph into an intermediate ast, where each key is a function
        # name
        # inputs
        # repeat
        # e.g. ("inputs", inputs_list, ("name", name_str, ("repeat", n, [])))
        ast:list = inter_graph_to_ast(_graph)

        # Step 3: first pass
        # A list of dicts
        # The reason of having a list of dicts as the main underlying data structure is due to its simplicity. This data structure can be transformed into the final ast
        # Each dict has the following structure:
        # "meta": {"inputs": None or list,
        #          "name": string (scope/.../name),
        #          "scope": the scope of the node,
        #          "type": type of the deep learning component}
        # "value": {"hyperparams": dict of hyperparameters,
        #           "stats": {}
        #           "tensor": library specific tensor object (not easily serializable),
        #           "value": a serializable object that contains the same info as tensor,
        #           "var": }
        graph:list = eval_ast_to_graph(ast)
        # Step 4: second pass: same data structure; populate inputs
        graph:list = global_update(graph)
        # if config_to_type(yml_file) not in const.ALL_RECIPES:
        graph = annotate(graph, lazy=lazy) # json schema
        # pprint(graph)
        # Step 5: load graph and states from checkpoint; check if the structure matches the one defined in the DSL
    except Exception:
        message = "-------------- Error during parsing configuration %s ----------" % yml_file
        traceback.print_exc()
        raise
    if return_dict:
        graph = list_to_dict(graph)
    return graph


## Util
def dict_to_arg_str(d):
    out = ""
    for key in d:
        if isinstance(d[key], str):
            val = "'%s'"%str(d[key])
        else:
            val = str(d[key])
        out += "%s=%s" % (str(key), val)
    return out


def make_graph_from_yml(yml_file, png_file, level, show="label", lazy=True):
    graph = parse(yml_file, lazy)
    draw_graph(graph, level, png_file, show=show)


def make_ast_from_yml(yml_file, png_file, lazy=True):
    graph = parse(yml_file, lazy)
    draw_hyperparam_tree(graph,
                         graph_path=png_file)



# ----------------------------------------------------------------------
# Created: lör maj 15 20:16:25 2021 (+0200)
# Last-Updated:
# Filename: test_annotator.py
# Author: Yinan Yu
# Description:
# ----------------------------------------------------------------------
import unittest
import os
from pprint import pprint
from alex.alex import core, const
from alex.annotators import param_count


class TestParamCount(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.config_path = "examples/configs/small2.yml"

    def test_annotate_trainbale_params(self):
        params_counter = param_count.ParamCount(self.config_path)
        annotated = params_counter.annotate_tree()
        core.draw(annotated,
                  graph_path=os.path.join(const.CACHE_BASE_PATH,
                                          'trainable_param_count.png'),
                  label_field="count",
                  excluded_types=["hyperparam"])
        core.draw(annotated,
                  graph_path=os.path.join(const.CACHE_BASE_PATH,
                                          'shape.png'),
                  label_field="shape",
                  excluded_types=["hyperparam"])


if __name__ == '__main__':
    unittest.main()



# ----------------------------------------------------------------------
# Created: ons mar  3 23:40:02 2021 (+0100)
# Last-Updated:
# Filename: ns_tf.py
# Author: Yinan Yu
# Description:
# ----------------------------------------------------------------------
from alex.alex import const
from alex.engine import ns_alex


ENGINE = "tf"
DIM_ORDER = ["batch_size", "height", "width", "channel"]


def add_imports(additional_modules=[]):
    default_modules = [["tensorflow", "tf"],
                       ["numpy", "np"]]
    modules = default_modules + additional_modules
    imports = ""
    for m in modules:
        imports += "import %s " % m[0]
        if len(m) == 2:
            imports += "as %s\n" % m[1]
        elif len(m) == 1:
            imports += "\n"
    imports += "\n\n"
    dtypes = "tf_dtypes = {'float32': tf.float32, 'int8': tf.int8}\n"
    space = "\n\n"
    return imports + dtypes + space


def add_global_configs():
    return ""


def wrap_in_class(trainable_params_code, component_code, loss_code, optimizer_code, *args, **kargs):
    code = []
    code.append(trainable_params_code)
    code.append(component_code)
    code.append(loss_code)
    code.append(optimizer_code)
    code = "\n\n".join(code)
    return code



import os
import queue
import random
import string
import threading
import time
import uuid
import yaml
import json
import numpy as np
from copy import deepcopy
from contextlib import contextmanager
from datetime import datetime
from glob import glob
import types
import io
from collections import OrderedDict, Iterable
import warnings
from copy import deepcopy
import shutil

from alex.alex import const


################## Data structure related
def flatten_subdict_keys(tree, parent=[], key_lst=[]):
    key_lst = deepcopy(key_lst)
    tree = deepcopy(tree)
    for k, v in tree.items():
        new_key = "/".join(parent+[k])
        if new_key in key_lst:
            duplicated_key = "Key %s exists" % new_key
            warnings.warn(duplicated_key)
        key_lst.append(new_key)
        if isinstance(v, dict):
            _, key_lst = flatten_subdict_keys(v, parent=[k], key_lst=key_lst)
    return parent, key_lst


def flatten_dict_keys(tree, ancestors=[], key_lst=[]):
    key_lst = deepcopy(key_lst)
    tree = deepcopy(tree)
    for k, v in tree.items():
        _ancestors = deepcopy(ancestors)
        new_key = "/".join(_ancestors+[k])
        if new_key in key_lst:
            duplicated_key = "Key %s exists" % new_key
            raise Exception(duplicated_key)
        key_lst.append(new_key)
        if isinstance(v, dict):
            _ancestors.append(k)
            _, key_lst = flatten_dict_keys(v, ancestors=_ancestors, key_lst=key_lst)
    return ancestors, key_lst


def keys_unique(tree):
    _, key_lst = flatten_subdict_keys(tree)
    keys = list(map(lambda x: x.split("/")[-1], key_lst))
    if len(keys) > len(list(set(keys))):
        return False
    else:
        return True


def fill_dict(dict1, dict2):
    # dict1 is new info, dict2 is original
    if not isinstance(dict2, dict):
        dict2 = dict()
    for k, v in dict1.items():
        if k not in dict2:
            dict2[k] = deepcopy(v)
        if isinstance(v, dict):
            fill_dict(v, dict2[k])


def update_ordereddict(new_dict, original_dict):
    for k, v in new_dict.items():
        if k not in original_dict:
            original_dict[k] = deepcopy(v)
        else:
            if isinstance(v, dict):
                update_ordereddict(v, original_dict[k])
            elif isinstance(v, list):
                original_dict[k] = list(set(deepcopy(v)).union(set(deepcopy(original_dict[k]))))
            elif isinstance(v, set):
                original_dict[k] = set(deepcopy(v)).union(set(deepcopy(original_dict[k])))
            else:
                original_dict[k] = deepcopy(v)


def replace_key(data, key, new_value):
    if isinstance(data, dict):
        if key in data:
            data[key] = deepcopy(new_value)
        for _key in data:
            if isinstance(data[_key], dict):
                if key in data[_key]:
                    data[_key][key] = deepcopy(new_value)
                else:
                    replace_key(data[_key], key, new_value)


def get_value(data, key, found=False):
    if not keys_unique(data):
        print("Warning: %s is not unique" %key)
    data = deepcopy(data)
    if found:
        return data, found
    for k, v in data.items():
        if k == key:
            return v, True
        if type(v) is dict:
            _v, found = get_value(v, key, found)
            if found:
                return _v, found
    return None, False


################## Arg checks
def arg_list(lst, exp_len):
    if not isinstance(lst, list):
        lst = [lst]
    if len(lst)!=exp_len:
        raise Exception("Input length not expected! %i!=%i" % (len(lst),
                                                               exp_len))
    return lst


################## File related
def get_component_path(component):
    return os.path.join(const.COMPONENT_BASE_PATH, component+".yml")


def get_component_spec_path(component):
    return os.path.join(const.COMPONENT_BASE_PATH, "."+component+".yml")


def read_yaml(yml_path):
    with open(yml_path, 'r') as params:
        try:
            config = yaml.load(params, Loader=yaml.SafeLoader)
        except yaml.YAMLError as exc:
            print(exc)
    return config


def read_json(json_path):
    with open(json_path) as _file:
        data = json.load(_file)
    return data


def generate_random_string(string_length):
    return '_' + (''.join(random.choice(string.ascii_lowercase
                                        + string.digits)
                          for _ in range(string_length)))


def get_latest_file(path, ext):
    return sorted(glob(os.path.join(path, "*"+ext)), key=os.path.getmtime)[-1]


def concatenate_files(filenames, outfile):
    lines = []
    for filename in filenames:
        with open(filename, "r") as stream:
            lines += stream.readlines()
        lines.append("\n")
        lines.append("\n")
    lines = "".join(lines)
    with open(outfile, "w") as stream:
        stream.write(lines)


def write_to(exp, filename, mode="a"):
    s = exp
    if filename is None:
        return s # TODO: write to IO
    else:
        with open(filename, mode) as stream:
            stream.write(s)


def clear_file(filename):
    open(filename, 'w').close()


def delete_dir(dir_name):
    try:
        shutil.rmtree(dir_name)
    except OSError as e:
        print("Error: %s - %s." % (e.filename, e.strerror))



# ----------------------------------------------------------------------
# Created: lör maj 22 10:39:51 2021 (+0200)
# Last-Updated:
# Filename: test_code_gen.py
# Author: Yinan Yu
# Description:
# ----------------------------------------------------------------------
import unittest
import os
from pprint import pprint
from alex.alex import core, const, dsl_parser, util
from alex.annotators import code_gen


class TestCodeGen(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.config_path = "examples/configs/small2.yml"
        self.config_path_alt = "examples/configs/small1.yml"
        self.boiler_plate = "alex/engine/example_boilerplate_tf.py" # "alex/engine/example_boilerplate_pytorch.py"

    def test_code_genenration(self):
        code_path = os.path.join(const.CACHE_BASE_PATH,
                                 "new_generated.py")
        code_generator = code_gen.CodeGen("new_generated.py",
                                          self.config_path,
                                          "tf",
                                          dirname=const.CACHE_BASE_PATH,
                                          load_ckpt=["cache",
                                                     None]) # "config_1622411488965296.json"
        # code_generator = code_gen.CodeGen("new_generated.py",
        #                                   self.config_path,
        #                                   "tf",
        #                                   dirname=const.CACHE_BASE_PATH,
        #                                   load_ckpt=["cache",
        #                                              "config_1622420349826577.json"]) # "config_1622411488965296.json"
        code_generator.generate_python()
        util.concatenate_files([code_path,
                                self.boiler_plate],
                               code_path)
        print("Saved to:", code_path)

    # def test_mismatched(self):
    #     code_path = os.path.join(const.CACHE_BASE_PATH,
    #                              "new_mismatched_generated.py")
    #     code_generator = code_gen.CodeGen("new_mismatched_generated.py",
    #                                       self.config_path_alt,
    #                                       "pytorch",
    #                                       dirname=const.CACHE_BASE_PATH,
    #                                       load_ckpt=["cache",
    #                                                  "config_1622420349826577.json"]) # "config_1622411488965296.json"
    #     annotated = code_generator.generate_python()
    #     util.concatenate_files([code_path,
    #                             "alex/engine/example_boilerplate_pytorch.py"],
    #                            code_path)
    #     print("Saved to:", code_path)


if __name__ == '__main__':
    unittest.main()



# ----------------------------------------------------------------------
# Created: mån maj 17 21:52:56 2021 (+0200)
# Last-Updated:
# Filename: param_count.py
# Author: Yinan Yu
# Description:
# ----------------------------------------------------------------------
import numpy as np
from math import floor, ceil
from pprint import pprint
import traceback
from copy import deepcopy
from alex.alex import core, const, node_interface
from alex.alex.annotator_interface import Annotator
from alex.engine import ns_alex


class Ingredient(node_interface.Ingredient):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.transformation = None

    def get_shape(self, node, *args, **kwargs):
        return node["meta"]["position"]["shape"]

    def get_trainable_params_count(self, node, annotated):
        name = node["name"]
        return sum(list(map(lambda x: x["count"],
                            core.get_children(name,
                                              annotated))))


class Recipe(node_interface.Recipe):

    def get_shape(self, node, annotated):
        last_component = annotated[node["children"][-1]]
        input_shape = last_component["meta"]["position"]["input_shape"][0]
        return input_shape

    def get_trainable_params_count(self, node, annotated):
        name = node["name"]
        return sum(list(map(lambda x: x["count"],
                            core.get_children(name,
                                              annotated))))


class Hyperparam(node_interface.Hyperparam):

    def get_trainable_params_count(self, node, annotated):
        if self.trainable:
            return np.prod(np.asarray(self.get_shape(node, annotated)))
        else:
            return 0

    def get_shape(self, *args, **kwargs):
        return [0, ]

    def get_input_shape(self, node):
        return node["meta"]["position"]["input_shape"]

# ------------------------------------------------------------------------ #
class Conv2DFilters(Hyperparam):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.trainable = True

    def get_shape(self, node, annotated):
        # input_shape = annotated[node["meta"]["position"]["component"]]["meta"]["position"]["input_shape"][0]
        input_shape = self.get_input_shape(node)[0]
        shape = node["meta"]["hyperparams"]["shape"]
        return [input_shape[-1],
                shape["kernel_size_h"],
                shape["kernel_size_w"],
                shape["n_filters"]]


class DenseWeights(Hyperparam):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.trainable = True

    def get_shape(self, node, annotated):
        # input_shape = annotated[node["meta"]["position"]["component"]]["meta"]["position"]["input_shape"][0]
        input_shape = self.get_input_shape(node)[0]
        return [input_shape[-1],
                node["meta"]["hyperparams"]["shape"]["n_units"]]


class DenseBias(Hyperparam):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.trainable = True

    def get_shape(self, *args, **kwargs):
        return [1, ]


class BatchnormMean(Hyperparam):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_shape(self, node, annotated):
        # input_shape = annotated[node["meta"]["position"]["component"]]["meta"]["position"]["input_shape"][0]
        input_shape = self.get_input_shape(node)[0]
        return [input_shape[-1], ]


class BatchnormVariance(Hyperparam):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_shape(self, node, annotated):
        # input_shape = annotated[node["meta"]["position"]["component"]]["meta"]["position"]["input_shape"][0]
        input_shape = self.get_input_shape(node)[0]
        return [input_shape[-1], ]


class BatchnormOffset(Hyperparam):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.trainable = True

    def get_shape(self, node, annotated):
        # input_shape = annotated[node["meta"]["position"]["component"]]["meta"]["position"]["input_shape"][0]
        input_shape = self.get_input_shape(node)[0]
        return [input_shape[-1], ]


class BatchnormScale(Hyperparam):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.trainable = True

    def get_shape(self, node, annotated):
        # input_shape = annotated[node["meta"]["position"]["component"]]["meta"]["position"]["input_shape"][0]
        input_shape = self.get_input_shape(node)[0]
        return [input_shape[-1], ]


def nodes(node):
    value = node["value"]
    # TODO: group ingredient by block
    nodes_ingredient = {"ingredient": Ingredient}
    nodes_hyperparam = {"hyperparam": Hyperparam}
    nodes_param = {"filters": Conv2DFilters,
                   "weights": DenseWeights,
                   "bias": DenseBias,
                   "means": BatchnormMean,
                   "vairance": BatchnormVariance,
                   "offset": BatchnormOffset,
                   "scale": BatchnormOffset}
    nodes_recipe = {"recipe": Recipe}

    _nodes = {**nodes_hyperparam,
              **nodes_ingredient,
              **nodes_param,
              **nodes_recipe}
    if value not in _nodes:
        value = node["type"]
    return _nodes[value](node["value"])


class ParamCount(Annotator):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.anno_name = "Trainable params count"
        self.passes = [[self.cache_shape,
                        self.annotate_count]]

    def annotate_count(self, node):
        node = deepcopy(node)
        node["count"] = nodes(node).get_trainable_params_count(node=node,
                                                               annotated=self.annotated)
        return node

    def cache_shape(self, node):
        """Cache the shape info in the node"""
        node = deepcopy(node)
        name = node["name"]

        node["shape"] = nodes(node).get_shape(node,
                                              self.annotated)
        input_nodes = self.get_input_nodes(name, self.annotated)
        if input_nodes is None:
            inputs = None
        else:
            inputs = list(map(lambda x: x["name"], input_nodes))
        node["input_nodes"] = inputs
        node["ancestor"] = core.get_ancestor_ingredient_node(node, self.tree)
        if node["ancestor"]["name"] in self.components:
            node["dtype"] = self.components[node["ancestor"]["name"]]["meta"]["dtype"]
        else:
            node["dtype"] = None
        return node



# ----------------------------------------------------------------------
# Created: tis mar 16 15:23:47 2021 (+0100)
# Last-Updated:
# Filename: engine_user.py
# Author: Yinan Yu
# Description:
# ----------------------------------------------------------------------
from alex.engine import ns_alex


def lr_scheduler(epoch, epochs, lrs):
    if len(lrs) != len(epochs)+1:
        raise Exception("Length mismatch!")
    prev_e = 0
    if epoch >= epochs[-1]:
        return lrs[-1]
    for i, l in enumerate(lrs):
        if prev_e <= epoch < epochs[i]:
            lr = l
            break
        prev_e = epochs[i]

    return lr



from tensorflow import keras
from tensorflow.keras import datasets
import matplotlib.pyplot as plt
import numpy as np

num_classes = 10
(x_train, label_train), (x_test, label_test) = datasets.cifar10.load_data()

# Normalize pixel values to be between 0 and 1
# Scale images to the [0, 1] range
x_train = x_train.astype("float32") / 255
x_test = x_test.astype("float32") / 255

# convert class vectors to binary class matrices
y_train = keras.utils.to_categorical(label_train, num_classes)
y_test = keras.utils.to_categorical(label_test, num_classes)


class_names = ['airplane', 'automobile', 'bird', 'cat', 'deer',
               'dog', 'frog', 'horse', 'ship', 'truck']

plt.figure(figsize=(10,10))
for i in range(25):
    plt.subplot(5,5,i+1)
    plt.xticks([])
    plt.yticks([])
    plt.grid(False)
    plt.imshow(x_train[i])
    # The CIFAR labels happen to be arrays,
    # which is why you need the extra index
    plt.xlabel(class_names[label_train[i][0]])
plt.show()


# from alex.alex.checkpoint import Checkpoint

# C = Checkpoint("examples/configs/small1.yml",
#                ["cache",  "config_1622420349826577.json"],
#                ["checkpoints", None])

# ckpt = C.load()

trainable_variables = get_trainable_params()

from alex.alex import const
var_list = []
for tv in trainable_variables:
    if tv.split("/")[-1] in const.ALL_TRAINABLE_PARAMS:
        var_list.append(trainable_variables[tv])


optimizer = get_optimizer(trainable_variables)


def train(x, gt, trainable_variables, var_list, optimizer):
    with tf.GradientTape() as tape:
        prediction = model(x, trainable_variables, training=True)
        gradients = tape.gradient(get_loss(trainable_variables, [gt, prediction]), var_list)
        optimizer.apply_gradients(zip(gradients, var_list))




num_epochs = 90
batch_size = 100

train_loss_results = []
train_accuracy_results = []

train_ds = tf.data.Dataset.from_tensor_slices(
    (x_train, y_train)).shuffle(50000).batch(batch_size)

test_ds = tf.data.Dataset.from_tensor_slices((x_test, y_test)).batch(10000)
for x_test, y_test in test_ds:
    break


for epoch in range(num_epochs):
    for i, (batch_x, batch_y) in enumerate(train_ds):
      train(batch_x, batch_y, trainable_variables, var_list, optimizer)

    preds = model(x_test, trainable_variables, training=False)

    matches_test  = tf.equal(tf.math.argmax(preds,1), tf.math.argmax(y_test,1))

    epoch_accuracy = tf.reduce_mean(tf.cast(matches_test,tf.float32))
    current_loss = get_loss(trainable_variables, [preds, y_test])
    epoch_loss_avg = tf.reduce_mean(current_loss)
    train_loss_results.append(epoch_loss_avg)
    train_accuracy_results.append(epoch_accuracy)

    tf.print("[", epoch, "500]", "accuracy: ", epoch_accuracy, ", loss: ",epoch_loss_avg)



# ----------------------------------------------------------------------
# Created: tis apr 20 10:06:05 2021 (+0200)
# Last-Updated:
# Filename: compare.py
# Author: Yinan Yu
# Description:
# ----------------------------------------------------------------------
from copy import deepcopy
from pprint import pprint
import numpy as np
import uuid
import matplotlib.pyplot as plt

from alex.alex import dsl_parser, core, const

# FIXME: POC
delete_cost = lambda x: 1
insert_cost = lambda x: 1
update_cost = lambda x, y: 1 if x!=y else 0
match_cost = lambda x: 0

DELETE = lambda x: ["DELETE", x, ""]
INSERT = lambda x: ["INSERT", "", x]
UPDATE = lambda x, y: ["UPDATE", x, y]
MATCH = lambda x, y: ["MATCH", x, y]


def get_lmds(tree):
    return list(map(lambda x: tree[x]["lmd"], tree))


def get_keyroots(tree):
    lmds = list(set(get_lmds(tree)))
    keyroots = {x: x for x in lmds}
    for node in tree:
        keyroots[tree[node]["lmd"]] = node
    return sorted(core.name_to_index(list(keyroots.values()), tree))


def ted(t1, t2):

    Al = core.name_to_index(get_lmds(t1), t1)
    Bl = core.name_to_index(get_lmds(t2), t2)
    An = core.get_labels(t1)
    Bn = core.get_labels(t2)

    Anames = list(t1.keys())
    Bnames = list(t2.keys())

    cost_matrix = np.zeros((len(An), len(Bn)))
    operations = [[[] for _ in range(len(Bn))] for _ in range(len(An))]

    keyroots1 = get_keyroots(t1)
    keyroots2 = get_keyroots(t2)
    for i in keyroots1:
        for j in keyroots2:
            _distance(i, j, Al, Bl, An, Bn, Anames, Bnames, cost_matrix, operations)
    return cost_matrix[-1,-1], operations[-1][-1]


def annotate_ops(operations):
    annotation = {1: dict(),
                  2: dict()}
    for operation in operations:
        node_name = operation[0]
        op = operation[1][0]
        if op == "DELETE":
            annotation[1][node_name] = "red"
        elif op == "INSERT":
            annotation[2][node_name] = "green"
        elif op == "UPDATE":
            annotation[1][node_name[0]] = "orange"
            annotation[2][node_name[1]] = "orange"
    return annotation


def label_to_value(label):
    return label.split("#")[0]


def _diff_graph_list(n1, n2, tree_full1, tree_full2):
    list1 = dsl_parser.list_to_graph(n1)
    list2 = dsl_parser.list_to_graph(n2)
    tree1 = core.alex_graph_to_tree(list1, exclude_types=["hyperparam"], naive=False)
    tree2 = core.alex_graph_to_tree(list2, exclude_types=["hyperparam"], naive=False)

    cost, _operations = ted(tree1, tree2)
    operations = []
    for operation in _operations:
        if operation[1][1] == "root":
            break
        op = operation[1][0]
        if op == "INSERT":
            operations.append(operation)
            label2 = operation[1][2]
            if core.get_value_type(label_to_value(label2)) == "ingredient":
                affected_hyperparam_subtree = tree_full2[operation[0]]["descendants"]
                for child in affected_hyperparam_subtree:
                    cost += 1 # FIXME
                    operations.append((child, ["INSERT", "", tree_full2[child]["label"]]))
        elif op == "DELETE":
            operations.append(operation)
            label1 = operation[1][1]
            if core.get_value_type(label_to_value(label1)) == "ingredient":
                affected_hyperparam_subtree = tree_full1[operation[0]]["descendants"]
                for child in affected_hyperparam_subtree:
                    cost += 1 # FIXME
                    operations.append((child, ["DELETE", tree_full1[child]["label"], ""]))
        else:
            name1 = operation[0][0]
            name2 = operation[0][1]
            label1 = operation[1][1]
            label2 = operation[1][2]
            if label1 != label2:
                if label_to_value(label1) == label_to_value(label2):
                    subtree1 = core.get_subtree(tree_full1, name1)
                    subtree2 = core.get_subtree(tree_full2, name2)
                    _cost, __operations = ted(subtree1,
                                              subtree2)
                    cost += _cost
                    operations += __operations[:-1]
                    # pprint(__operations[:-1])
                else:
                    operations.append((name1, ("DELETE", label1, "")))
                    operations.append((name2, ("INSERT", "", label2)))
                    if core.get_value_type(label_to_value(label1)) == "ingredient":

                        affected_hyperparam_subtree = tree_full1[name1]["descendants"]
                        for child in affected_hyperparam_subtree:
                            cost += 1 # FIXME
                            _operation = (child, ["DELETE", tree_full1[child]["label"], ""])
                            operations.append(_operation)
                    if core.get_value_type(label_to_value(label2)) == "ingredient":
                        affected_hyperparam_subtree = tree_full2[name2]["descendants"]
                        for child in affected_hyperparam_subtree:
                            cost += 1 # FIXME
                            _operation = (child, ["INSERT", "", tree_full2[child]["label"]])
                            operations.append(_operation)

    return cost, operations


def _distance(i, j, Al, Bl, An, Bn, Anames, Bnames, cost_matrix, operations):
    """ Disclaimer: the _distance function is modified from the equivalent
            method implemented in the python library zss:
            https://github.com/timtadh/zhang-shasha/blob/master/zss/compare.py
    """
    m = i - Al[i] + 2
    n = j - Bl[j] + 2

    fd = np.zeros((m, n))
    partial_ops = [[[] for _ in range(n)] for _ in range(m)]

    ioff = Al[i] - 1
    joff = Bl[j] - 1

    for x in range(1, m): # δ(l(i1)..i, θ) = δ(l(1i)..1-1, θ) + γ(v → λ)
        node = An[x+ioff]
        nameA = Anames[x+ioff]
        fd[x, 0] = fd[x-1, 0] + delete_cost(node)
        op = (nameA, DELETE(node))
        partial_ops[x][0] = partial_ops[x-1][0] + [op]
    for y in range(1, n): # δ(θ, l(j1)..j) = δ(θ, l(j1)..j-1) + γ(λ → w)
        node = Bn[y+joff]
        nameB = Bnames[y+joff]
        fd[0,y] = fd[0,y-1] + insert_cost(node)
        op = (nameB, INSERT(node))
        partial_ops[0][y] = partial_ops[0][y-1] + [op]

    for x in range(1, m):  # the plus one is for the xrange impl
        for y in range(1, n):
            # x+ioff in the fd table corresponds to the same node as x in
            # the cost_matrix table (same for y and y+joff)
            node1 = An[x+ioff]
            node2 = Bn[y+joff]
            node_name1 = Anames[x+ioff]
            node_name2 = Bnames[y+joff]
            # only need to check if x is an ancestor of i
            # and y is an ancestor of j

            if Al[i] == Al[x+ioff] and Bl[j] == Bl[y+joff]:
                #                   +-
                #                   | δ(l(i1)..i-1, l(j1)..j) + γ(v → λ)
                # δ(F1 , F2 ) = min-+ δ(l(i1)..i , l(j1)..j-1) + γ(λ → w)
                #                   | δ(l(i1)..i-1, l(j1)..j-1) + γ(v → w)
                #                   +-
                costs = [fd[x-1, y] + delete_cost(node1),
                         fd[x,y-1] + insert_cost(node2),
                         fd[x-1,y-1] + update_cost(node1, node2)]
                fd[x,y] = min(costs)
                min_index = costs.index(fd[x,y])
                if min_index == 0:
                    op = (node_name1, DELETE(node1))
                    partial_ops[x][y] = partial_ops[x-1][y] + [op]
                elif min_index == 1:
                    op = (node_name2, INSERT(node2))
                    partial_ops[x][y] = partial_ops[x][y - 1] + [op]
                else:
                    op_type = MATCH if fd[x,y] == fd[x-1,y-1] else UPDATE
                    op = ((node_name1, node_name2), op_type(node1, node2))

                    partial_ops[x][y] = partial_ops[x - 1][y - 1] + [op]

                operations[x + ioff][y + joff] = partial_ops[x][y]
                cost_matrix[x+ioff,y+joff] = fd[x,y]
            else:
                #                   +-
                #                   | δ(l(i1)..i-1, l(j1)..j) + γ(v → λ)
                # δ(F1 , F2 ) = min-+ δ(l(i1)..i , l(j1)..j-1) + γ(λ → w)
                #                   | δ(l(i1)..l(i)-1, l(j1)..l(j)-1)
                #                   |                     + treedist(i1,j1)
                #                   +-
                p = Al[x+ioff]-1-ioff
                q = Bl[y+joff]-1-joff
                costs = [fd[x-1,y] + delete_cost(node1),
                         fd[x,y-1] + insert_cost(node2),
                         fd[p,q] + cost_matrix[x+ioff,y+joff]]
                fd[x,y] = min(costs)
                min_index = costs.index(fd[x,y])
                if min_index == 0:
                    op = (node_name1, DELETE(node1))
                    partial_ops[x][y] = partial_ops[x-1][y] + [op]
                elif min_index == 1:
                    op = (node_name2, INSERT(node2))
                    partial_ops[x][y] = partial_ops[x][y-1] + [op]
                else:
                    partial_ops[x][y] = partial_ops[p][q] + \
                        operations[x+ioff][y+joff]


def _dist_graph_list(tree1, tree2, exclude_types=[], render_to=None, dpi=800):
    cost, operations = ted(tree1, tree2)
    if render_to is not None:
        print("Done computing distance. Rendering image")
        annotation = annotate_ops(operations)
        img1 = core.draw(tree1, None, annotation[1], dpi=dpi)
        img2 = core.draw(tree2, None, annotation[2], dpi=dpi)
        fig, axs = plt.subplots(1, 2, dpi=dpi)
        axs[0].imshow(img1)
        axs[0].axis("off")
        axs[1].imshow(img2)
        axs[1].axis("off")
        fig.tight_layout()
        if render_to != "":
            fig.savefig(render_to, dpi="figure")
            print("Distance images written to: %s" % render_to)
        else:
            print("Distance images rendered to screen")
            plt.show()
    return cost, operations


from alex.annotators import param_count

def diff(network_config_1,
         network_config_2,
         render_to=None,
         dpi=800, lazy=False):
    graph_list1 = dsl_parser.parse(network_config_1, lazy=lazy)
    graph_list2 = dsl_parser.parse(network_config_2, lazy=lazy)
    # Do not include dimensionality in diff
    # graph1 = dsl_parser.list_to_graph(graph_list1)
    # graph2 = dsl_parser.list_to_graph(graph_list2)
    # tree_full1 = core.alex_graph_to_tree(graph1, naive=False)
    # tree_full2 = core.alex_graph_to_tree(graph2, naive=False)
    # Include dimensionality in diff
    tree_full1 = param_count.ParamCount(graph_list1).annotate_tree()
    tree_full2 = param_count.ParamCount(graph_list2).annotate_tree()

    operations = []
    cost = 0
    for block in ["data", "model", "loss", "optimizer"]:

        graph_list1_block = list(filter(lambda x: x["meta"]["block"]==block, graph_list1))
        graph_list2_block = list(filter(lambda x: x["meta"]["block"]==block, graph_list2))
        # for component in graph_list1_block:
        #     name = component["meta"]["name"]
        #     component["meta"]["shape"] = tree_full1[name]["shape"]
        # for component in graph_list2_block:
        #     name = component["meta"]["name"]
        #     component["meta"]["shape"] = tree_full2[name]["shape"]

        _cost, _operations = _diff_graph_list(graph_list1_block,
                                              graph_list2_block,
                                              tree_full1, tree_full2)
        cost += _cost
        operations += _operations

    if render_to is not None:
        print("Done computing diff. Rendering image")
        annotation = annotate_ops(operations)
        img1 = core.draw(tree_full1, None, annotation[1], dpi=dpi)
        img2 = core.draw(tree_full2, None, annotation[2], dpi=dpi)
        fig, axs = plt.subplots(1, 2, dpi=dpi)
        axs[0].imshow(img1)
        axs[0].axis("off")
        axs[1].imshow(img2)
        axs[1].axis("off")
        fig.tight_layout()
        if render_to != "":
            fig.savefig(render_to, dpi="figure")
            print("Diff images written to: %s" % render_to)
        else:
            print("Diff images rendered to screen")
            plt.show()

    return cost, operations


def dist(network_config_1,
         network_config_2,
         render_to=None,
         exclude_types=[],
         dpi=800, lazy=False):
    graph_list1 = dsl_parser.parse(network_config_1, lazy=lazy)
    graph_list2 = dsl_parser.parse(network_config_2, lazy=lazy)
    # Do not include dimensionality in dist
    # graph1 = dsl_parser.list_to_graph(graph_list1)
    # graph2 = dsl_parser.list_to_graph(graph_list2)
    # tree1 = core.alex_graph_to_tree(graph1, exclude_types=exclude_types)
    # tree2 = core.alex_graph_to_tree(graph2, exclude_types=exclude_types)
    # Include dimensionality in dist
    tree1 = param_count.ParamCount(graph_list1).annotate_tree()
    tree2 = param_count.ParamCount(graph_list2).annotate_tree()

    cost, operations = _dist_graph_list(tree1,
                                        tree2,
                                        exclude_types=exclude_types,
                                        render_to=render_to,
                                        dpi=dpi)
    return cost, operations


# ---------------------------------------------------------------------------- #
# Replace subtree
# ---------------------------------------------------------------------------- #
# - _, operation = dist(network1, network2, exclude_types=["hyperparam"])
# Find all matched ingredients
# Load checkpoint
def matched_ingredients(network_config_1,
                        network_config_2,
                        render_to=None,
                        dpi=800):
    # n1 = dsl_parser.list_to_graph(graph_list1)
    # n2 = dsl_parser.list_to_graph(graph_list2)
    # tree1 = core.alex_graph_to_tree(n1, exclude_types=["hyperparam"], naive=False)
    # tree2 = core.alex_graph_to_tree(n2, exclude_types=["hyperparam"], naive=False)
    tree1 = param_count.ParamCount(network_config_1,
                                   naive=False).annotate_tree()
    tree2 = param_count.ParamCount(network_config_2,
                                   naive=False).annotate_tree()
    _, operations = ted(tree1, tree2)
    # pprint(operations)
    if render_to is not None:
        print("Done computing diff. Rendering image")
        annotation = annotate_ops(operations)
        img1 = core.draw(tree1, None, annotation[1], dpi=dpi)
        img2 = core.draw(tree2, None, annotation[2], dpi=dpi)
        fig, axs = plt.subplots(1, 2, dpi=dpi)
        axs[0].imshow(img1)
        axs[0].axis("off")
        axs[1].imshow(img2)
        axs[1].axis("off")
        fig.tight_layout()
        if render_to != "":
            fig.savefig(render_to, dpi="figure")
            print("Diff images written to: %s" % render_to)
        else:
            print("Diff images rendered to screen")
            plt.show()
    matched = {}
    for operation in operations:
        if operation[1][0] == "MATCH":
            component_type = label_to_value(operation[1][1])
            if component_type in const.PARAMS:
                matched = {**matched,
                           **{operation[0][0]:
                              operation[0][1]}}
    return matched



# ----------------------------------------------------------------------
# Created: fre maj 21 20:16:37 2021 (+0200)
# Last-Updated:
# Filename: code_gen.py
# Author: Yinan Yu
# Description:
# Steps for code generation:
# Step 1: generate translation code
# e.g.
# def xavier_uniform(seed, shape, dtype, name):
#    return torch.nn.init.xavier_uniform_(tensor=torch.empty(*shape), dtype=torch_types[dtype])
# Step 2: Traverse tree; generate function calls
# e.g.
# xavier_uniform_id = xavier_uniform(name="some_name", seed=1)
# Step 3: Nomralize function calls
# e.g.
# xavier_uniform_id = xavier_uniform(name="some_name", seed=1, dtype="float32", shape=(64, 3, 3, 3))
# Step 4: Inline
# ----------------------------------------------------------------------
import numpy as np
import os
import collections
from math import floor
from pprint import pprint
from copy import deepcopy
from typing import TypeVar, Union
import rope
from rope.base.project import Project
from rope.refactor.inline import create_inline
import warnings
import traceback

from alex.alex import core, const, node_interface, checkpoint, util
from alex.alex.checkpoint import Checkpoint
from alex.alex.annotator_interface import Annotator
from alex.engine import ns_alex, ns_tf, ns_pytorch
from alex.annotators import param_count

# TODO:
# [ ] Modify annotation (node_type, block, etc)
# [x] Add checkpoint
# [ ] Refine user interface
# [ ] Cache shape info during tree construction so we can also check the size at "compile" time

NAMESPACES = {"tf": ns_tf,
              "pytorch": ns_pytorch}


VALUE = TypeVar('VALUE', str, list, int, float)
FUNCTION = TypeVar('FUNCTION')
IDENTIFIER = TypeVar('IDENTIFIER')

special_keys = ["channels", "batch_size"]


def get_node_type(node):
    if node["value"] in ["input_shape"]:
        return IDENTIFIER # FIXME: temporary for backward compatibility
    elif node["value"] in const.FNS + ["shape"] + const.ALL_PARAMS_LIST:
        return FUNCTION
    elif len(node["children"])==0:
        return VALUE
    elif len(node["children"]) == len(node["descendants"]):
        return IDENTIFIER
    else:
        return IDENTIFIER


def _name_strs_to_names(name_strs):
    need_unpack = False
    if isinstance(name_strs, str):
        name_strs = [name_strs]
        need_unpack = True
    names = list(map(lambda x: x.replace("/", "_").replace("-", "_"),
                     name_strs))
    if need_unpack:
        names = names[0]
    return names


def _parse_str(value, node=None, literal=True):
    if not isinstance(value, str):
        return str(value)
    if value in special_keys:
        if value == "channels":
            parsed = node["input_shape"][-1]
        elif value == "batch_size":
            parsed = "batch_size"
            literal = False
    if value in const.ALL_COMPONENTS:
        parsed = "%s()" % value
    elif literal:
        parsed = "'%s'" % value
    else:
        parsed = "%s" % value
    return parsed


def value_to_kwarg_str(key, value, node=None, literal=True):
    if isinstance(value, list) and len(value)==1:
        value = value[0]
    if key != "":
        key = "%s=" % key
    if isinstance(value, str):
        value = _parse_str(value, node=node, literal=literal)
        kwarg_str = "%s%s" % (key, value)
    elif isinstance(value, list):
        _kwarg_str = []
        for _value in value:
            _kwarg_str.append(_parse_str(_value,
                                         node=node,
                                         literal=literal))
        _kwarg_str = ", ".join(_kwarg_str)
        kwarg_str = "%s[%s]" % (key, _kwarg_str)
    else:
        kwarg_str = "%s%s" % (key, value)
    return kwarg_str


def cache_fn(node: dict,
             fn: str,
             args: dict,
             return_symbol: str,
             name: str,
             node_type,
             inline: bool = False,
             extra: Union[str, None]=None):
    """
    return_symbol = fn(args["key"]=args["value"], ...)
    """
    node = deepcopy(node)
    if "code" not in node:
        node["code"] = collections.OrderedDict()

    arg_str = []
    for _arg in args:
        arg = args[_arg]
        if arg["value"] is not None:
            if arg["type"] == VALUE:
                _arg_str = value_to_kwarg_str("",
                                              arg["value"])
            elif arg["type"] == IDENTIFIER:
                if node_type == FUNCTION:
                    if arg["str"] == "":
                        continue
                    _arg_str = "%s=%s" % (arg["key"], arg["str"])
                else:
                    if arg["value"] == "":
                        continue
                    _arg_str = arg["value"]
            else:
                if arg["value"] == "":
                    continue
                if node_type == FUNCTION:
                    _arg_str = "%s=%s" % (arg["key"], arg["value"])
                else:
                    _arg_str = arg["value"]

            arg_str.append(_arg_str)

    if node_type == IDENTIFIER and len(arg_str)>1:
        arg_str = "[%s]" % ", ".join(arg_str)
    else:
        arg_str = ", ".join(arg_str)
    if node_type == FUNCTION:
        if len(arg_str) != "":
            arg_str = "(%s)" % (arg_str)
        code_str = "%s%s" % (fn,
                             arg_str)
    else:
        code_str = arg_str
    node["code"] = {"fn": fn,
                    "args": args,
                    "return_symbol": return_symbol,
                    "type": node_type,
                    "name": name,
                    "str": code_str,
                    "inline": inline,
                    "extra": extra}
    return node


class Ingredient(param_count.Ingredient):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def generate_code(self, node, annotated, engine):
        children = core.get_children(node["name"], annotated)
        args = dict()
        if node["input_nodes"] is not None:
            inputs = _name_strs_to_names(node["input_nodes"])
            if len(inputs) == 1:
                inputs = inputs[0]
        else:
            inputs = None

        args[const.ALEX_ARG_INPUTS] = {"key": const.ALEX_ARG_INPUTS,
                                       "value": inputs,
                                       "ref": inputs,
                                       "type": IDENTIFIER,
                                       "str": value_to_kwarg_str("", inputs,
                                                                 literal=False)}
        args[const.ALEX_ARG_NAME] = {"key": const.ALEX_ARG_NAME,
                                     "value": _name_strs_to_names(node["name"]),
                                     "ref": None,
                                     "type": IDENTIFIER,
                                     "str": "'%s'" % (node["name"])}
        if node["input_nodes"] is not None:
            input_nodes = annotated[node["input_nodes"][0]]
            args["shape"] = {"key": "shape",
                             "value": input_nodes,
                             "ref": None,
                             "type": IDENTIFIER,
                             "str": "%s" % str(input_nodes["shape"])}
        args["dtype"] = {"key": "dtype",
                         "value": node["dtype"],
                         "ref": None,
                         "type": IDENTIFIER,
                         "str": "'%s'" % (node["dtype"])}
        return_symbol = _name_strs_to_names(node["name"])

        for child in children:
            if child["value"] in args:
                continue
            _arg = child["code"]["args"]
            for _key in _arg:
                if _arg[_key]["type"] == IDENTIFIER:
                    args = {**args, **{_key: _arg[_key]}}
            if child["code"]["type"] != VALUE:
                if child["value"] in const.ALL_PARAMS:
                    _name = "%s/%s" % (core.get_parent(child["name"],
                                                       annotated,
                                                       "name"),
                                       child["value"])

                    args = {**args,
                            **{child["value"]: {"key": child["value"],
                                                "value": "trainable_params['%s']" % _name,
                                                "ref": child["name"],
                                                "type": child["code"]["type"],
                                                "str": child["code"]["str"]}}}
                else:
                    _child = {"key": child["value"],
                              "value": child["code"]["return_symbol"],
                              "ref": child["name"],
                              "type": child["code"]["type"],
                              "str": child["code"]["str"]}
                    if child["value"]=="decay":
                        if engine == "tf":
                            _child["key"] = "learning_rate"
                            args["learning_rate"] = _child
                    else:
                        args = {**args,
                                **{child["value"]: _child}}
        node = cache_fn(node,
                        node["value"],
                        args,
                        return_symbol,
                        node["name"],
                        FUNCTION)
        return node


class Regularizer(param_count.Ingredient):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def generate_code(self, node, annotated, engine):
        children = core.get_children(node["name"], annotated)
        args = dict()
        args[const.ALEX_ARG_INPUTS] = {"key": const.ALEX_ARG_INPUTS,
                                       "value": node["meta"]["position"]["inputs"],
                                       "ref": None,
                                       "type": IDENTIFIER,
                                       "str": value_to_kwarg_str("",
                                                                 node["meta"]["position"]["inputs"],
                                                                 literal=True)}
        args[const.ALEX_ARG_NAME] = {"key": const.ALEX_ARG_NAME,
                                     "value": _name_strs_to_names(node["name"]),
                                     "ref": None,
                                     "type": IDENTIFIER,
                                     "str": "'%s'" % (node["name"])}
        args["dtype"] = {"key": "dtype",
                         "value": node["dtype"],
                         "ref": None,
                         "type": IDENTIFIER,
                         "str": "'%s'" % (node["dtype"])}
        return_symbol = _name_strs_to_names(node["name"])

        for child in children:
            _arg = child["code"]["args"]
            for _key in _arg:
                if _arg[_key]["type"] == IDENTIFIER:
                    args = {**args, **{_key: _arg[_key]}}
            if child["code"]["type"] != VALUE:
                if child["value"] in const.ALL_PARAMS:
                    _name = "%s/%s" % (core.get_parent(child["name"],
                                                       annotated,
                                                       "name"),
                                       child["value"])

                    args = {**args,
                            **{child["value"]: {"key": child["value"],
                                                "value": "trainable_params['%s']" % _name,
                                                "ref": child["name"],
                                                "type": child["code"]["type"],
                                                "str": child["code"]["str"]}}}
                else:
                    args = {**args,
                            **{child["value"]: {"key": child["value"],
                                                "value": child["code"]["return_symbol"],
                                                "ref": child["name"],
                                                "type": child["code"]["type"],
                                                "str": child["code"]["str"]}}}
        node = cache_fn(node,
                        node["value"],
                        args,
                        return_symbol,
                        node["name"],
                        FUNCTION)
        return node


class Recipe(param_count.Recipe):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def generate_code(self, node, annotated, engine):
        node = deepcopy(node)
        node["code"] = dict()
        node["code"]["type"] = IDENTIFIER
        return node


class Function(param_count.Hyperparam):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def generate_code(self, node, annotated, engine):
        node = deepcopy(node)
        args = dict()
        return_symbol = _name_strs_to_names(node["name"])
        children = core.get_children(node["name"], annotated)
        for child in children:
            _node_type = child["code"]["type"]
            if _node_type == VALUE:
                continue
            elif _node_type == IDENTIFIER:
                _value = child["code"]["str"]
            elif _node_type == FUNCTION:
                _value = child["code"]["return_symbol"]
            args = {**args,
                    **{child["value"]: {"key": child["value"],
                                        "value": _value,
                                        "ref": child["name"],
                                        "type": child["code"]["type"],
                                        "str": child["code"]["str"]}}}
        node = cache_fn(node,
                        node["value"],
                        args,
                        return_symbol,
                        "",
                        FUNCTION)
        return node


class Value(param_count.Hyperparam):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def generate_code(self, node, annotated, engine):
        if "padding" in node["parent"] and engine=="pytorch":
            ingredient = annotated[node["meta"]["position"]["component"]]
            if ingredient["value"] in ["conv", "max_pool2d"]:
                if node["value"] == "SAME":
                    _value = "same"
                    # _value = [1, 1]
                    # inputs = annotated[ingredient["meta"]["position"]["inputs"][0]]
                    # input_shape = inputs["shape"]
                    # h, w = input_shape[0], input_shape[1]
                    strides = ingredient["meta"]["hyperparams"]["strides"]
                    if strides[0]!=1 or strides[1]!=1:
                        raise Exception("Pytorch does not support strided ops  for padding='same'")
                    # if ingredient["value"] == "conv":
                    #     k_h = ingredient["meta"]["hyperparams"]["filters"]["shape"]["kernel_size_h"]
                    #     k_w = ingredient["meta"]["hyperparams"]["filters"]["shape"]["kernel_size_w"]
                    # else:
                    #     [k_h, k_w] = ingredient["meta"]["hyperparams"]["window_shape"]
                    # padding_h_min = floor(((ingredient["shape"][0]-1)*strides[0] - h + k_h)/2)
                    # padding_w_min = floor(((ingredient["shape"][1]-1)*strides[1] - w + k_w)/2)
                    # _value = [padding_h_min, padding_w_min]
                elif node["value"] == "VALID":
                    # _value = [0, 0]
                    _value = "valid"
        else:
            _value = node["value"]
        node = cache_fn(node=deepcopy(node),
                        fn=_value,
                        args={},
                        return_symbol=_value,
                        name="",
                        node_type=VALUE)
        return node


class Shape(param_count.Hyperparam):

    def generate_code(self, node, annotated, engine):
        node = deepcopy(node)
        node_type = get_node_type(node)
        # FIXME: refactor dispatch; for now it's experimental
        args = dict()
        fn = "%s_%s_shape" % (node["ancestor"]["value"],
                              core.get_parent(node["name"],
                                              annotated,
                                              "value"))
        return_symbol = _name_strs_to_names(node["name"])

        children = core.get_children(node["name"], annotated)
        for child in children:
            _node_type = child["code"]["type"]
            if _node_type == VALUE:
                continue
            elif _node_type == IDENTIFIER:
                _value = child["code"]["str"]
            elif _node_type == FUNCTION:
                _value = child["code"]["return_symbol"]
            args = {**args,
                    **{child["value"]: {"key": child["value"],
                                        "value": _value,
                                        "ref": child["name"],
                                        "type": child["code"]["type"],
                                        "str": child["code"]["str"]}}}
        node = cache_fn(node,
                        fn,
                        args,
                        return_symbol,
                        "",
                        node_type,
                        True)
        return node


class Identifier(param_count.Hyperparam):

    def generate_code(self, node, annotated, engine):
        node = deepcopy(node)
        node_type = get_node_type(node)

        children = core.get_children(node["name"], annotated)

        args = dict()
        for child in children:
            # FIXME:
            if "code" not in child:
                continue
            args[child["value"]] = {"key": child["value"],
                                    "value": child["code"]["return_symbol"],
                                    "ref": child["name"],
                                    "type": child["code"]["type"],
                                    "str": child["code"]["str"]}
        node = cache_fn(node=node,
                        fn=node["value"],
                        args=args,
                        return_symbol=_name_strs_to_names(node["name"]),
                        name="",
                        node_type=node_type)
        return node


class Initializer(param_count.Hyperparam):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def generate_code(self, node, annotated, engine):
        node = deepcopy(node)
        node_type = get_node_type(node)

        args = dict()

        param_name = core.get_ancestor_param_node(node, annotated, field="value")
        shape_name = _name_strs_to_names("%s_%s_shape" % (node["ancestor"]["name"],
                                                          param_name))

        _name = node["name"]
        return_symbol = _name_strs_to_names(_name)

        args["shape"] = {"key": "shape",
                         "value": shape_name,
                         "ref": None,
                         "type": IDENTIFIER,
                         "str": "%s" % shape_name}

        args[const.ALEX_ARG_NAME] = {"key": const.ALEX_ARG_NAME,
                                     "value": _name_strs_to_names(_name),
                                     "ref": None,
                                     "type": IDENTIFIER,
                                     "str": "'%s'" % (_name)}

        children = core.get_children(node["name"], annotated)
        for child in children:
            _node_type = child["code"]["type"]
            if _node_type == VALUE:
                continue
            elif _node_type == IDENTIFIER:
                _value = child["code"]["str"]
            elif _node_type == FUNCTION:
                _value = child["code"]["return_symbol"]
            args = {**args,
                    **{child["value"]: {"key": child["value"],
                                        "value": _value,
                                        "ref": child["name"],
                                        "type": child["code"]["type"],
                                        "str": child["code"]["str"]}}}
        fn = node["value"]
        node = cache_fn(node,
                        fn,
                        args,
                        return_symbol,
                        _name,
                        node_type)

        return node


class TrainableParams(param_count.Hyperparam):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def generate_code(self, node, annotated, engine):
        node = deepcopy(node)
        node_type = get_node_type(node)
        args = dict()
        _name = "%s/%s" % (core.get_parent(node["name"],
                                           annotated,
                                           "name"),
                           node["value"])
        return_symbol = _name_strs_to_names(_name)
        node_type = FUNCTION

        ingredient = core.get_parent(node["name"],
                                     annotated,
                                     "value")
        is_trainable = const.PARAMS[ingredient][node["value"]]["derivative"]
        args["is_trainable"] = {"key": "is_trainable",
                                "value": is_trainable,
                                "ref": None,
                                "type": IDENTIFIER,
                                "str": "%s" % str(is_trainable)}
        fn = "%s_%s" % (core.get_parent(node["name"],
                                        annotated,
                                        "value"),
                        node["value"])

        args[const.ALEX_ARG_NAME] = {"key": const.ALEX_ARG_NAME,
                                     "value": _name_strs_to_names(_name),
                                     "ref": None,
                                     "type": IDENTIFIER,
                                     "str": "'%s'" % (_name)}
        children = core.get_children(node["name"], annotated)
        for child in children:
            if "code" not in child:
                continue
            _node_type = child["code"]["type"]
            if _node_type == VALUE:
                continue
            elif _node_type == IDENTIFIER:
                _value = child["code"]["str"]
            elif _node_type == FUNCTION:
                _value = child["code"]["return_symbol"]
            args = {**args,
                    **{child["value"]: {"key": child["value"],
                                        "value": _value,
                                        "ref": child["name"],
                                        "type": child["code"]["type"],
                                        "str": child["code"]["str"]}}}
        node = cache_fn(node=node,
                        fn=fn,
                        args=args,
                        return_symbol=return_symbol,
                        name=_name,
                        node_type=node_type,
                        inline=False,
                        extra="%s['%s'] = %s\n" % (const.ALEX_ARG_TRAINABLE_PARAMS,
                                                   _name,
                                                   return_symbol))
        return node


def nodes(node):
    value = node["value"]
    if value in const.REGULARIZERS:
        value = "regularizer"
    elif value in const.ALL_INITIALIZERS:
        value = "initializer_fn"
    elif node["value"] in const.ALL_PARAMS:
        value = "trainable_params"
    elif node["value"] in const.INGREDIENT_TYPES:
        value = "ingredient"
    elif node["value"] in const.SCHEDULER_BLOCK:
        value = FUNCTION
    _nodes = {"ingredient": Ingredient,
              "recipe": Recipe,
              "initializer_fn": Initializer,
              "trainable_params": TrainableParams,
              "shape": Shape,
              "regularizer": Regularizer,
              FUNCTION: Function,
              VALUE: Value,
              IDENTIFIER: Identifier}
    if value not in _nodes:
        if node["type"] == "hyperparam":
            value = get_node_type(node)
        else:
            value = node["type"]
    return _nodes[value](node["value"])


class CodeGen(param_count.ParamCount):

    def __init__(self,
                 output_file,
                 config_path,
                 engine,
                 dirname=const.CACHE_BASE_PATH,
                 load_ckpt=[const.CACHE_BASE_PATH, None],
                 save_ckpt=[const.CACHE_BASE_PATH, None]):
        super().__init__(config_path)

        if load_ckpt[1] is not None:
            self.load = True
        else:
            self.load = False

        self.ckpt = checkpoint.Checkpoint(config_path,
                                          load_ckpt,
                                          save_ckpt)
        self.anno_name = "code generation"
        self.engine = engine
        self.dirname = dirname
        self.alex_cache_code_path = os.path.join(dirname, "python_code_cache/")
        os.makedirs(self.alex_cache_code_path, exist_ok=True)

        self.cache_def_path = os.path.join(self.alex_cache_code_path, ".def.py")
        self.output_file = os.path.join(dirname, output_file)
        util.clear_file(self.cache_def_path)

        self.passes = [[self.cache_shape],
                       [self.generate_alex]]
        self.inline_index_translation = []
        self.inline_index_python = []
        self.blocks = {"param": {**const.ALL_PARAMS, **const.ALL_INITIALIZERS},
                       "model": const.MODEL_BLOCK,
                       "optimizer": const.OPTIMIZER_BLOCK,
                       "loss": const.LOSS_BLOCK,
                       "scheduler": const.SCHEDULER_BLOCK,
                       "data": const.INPUT_TYPES}
        if engine == "tf":
            self.blocks["optimizer"] = {**self.blocks["optimizer"],
                                        **self.blocks["scheduler"]}
            self.blocks["scheduler"] = {}
        self.filepaths = {"param": os.path.join(self.alex_cache_code_path, "_param_.py"),
                          "model": os.path.join(self.alex_cache_code_path, "_component_.py"),
                          "optimizer": os.path.join(self.alex_cache_code_path, "_optimizer_.py"),
                          "loss": os.path.join(self.alex_cache_code_path, "_loss_.py"),
                          "scheduler": os.path.join(self.alex_cache_code_path, "_scheduler_.py"),
                          "data": ""
        }
        self.alex_defs = {"param": [],
                          "model": [],
                          "optimizer": [],
                          "scheduler": [],
                          "loss": [],
                          "data": []}
        self.alex_code = {"param": [],
                          "model": [],
                          "optimizer": [],
                          "scheduler": [],
                          "loss": [],
                          "data": []}
        self.alex_inline = {"param": [],
                            "model": [],
                            "optimizer": [],
                            "scheduler": [],
                            "loss": [],
                            "data": []}

        for block in self.blocks:
            if block == "param":
                self.cache_param_translation()
            elif block == "data":
                continue
            self.cache_translation(block)
            write_list_to_file(self.alex_defs[block],
                               self.cache_def_path)
            util.clear_file(self.filepaths[block])
        util.clear_file(self.output_file)

    # FIXME: fix get_block logic
    def get_block(self, node):
        fn = node["value"]
        if "block" in node["meta"] and node["meta"]["block"] is not None:
            block = node["meta"]["block"]
        else:
            block = list(filter(lambda x: fn in self.blocks[x],
                                self.blocks))
            if len(block) != 0:
                block = block[0]
            else:

                _component_ = core.get_ancestor_ingredient_node(node,
                                                                self.annotated,
                                                                "value")
                _param_ = core.get_ancestor_param_node(node,
                                                       self.annotated,
                                                       "value")
                if _param_ is not None:
                    fn = _param_
                elif _component_ is not None:
                    fn = _component_
                block = list(filter(lambda x: fn in self.blocks[x],
                                    self.blocks))[0]
        return block

    @staticmethod
    def get_trg_fn(component, engine):
        return component[engine][0]

    @staticmethod
    def get_trg_args_str(component, engine):
        trg_args = list(map(lambda x: "%s=%s" % (x,
                                                 component[engine][1][x]),
                            component[engine][1]))
        return "(%s)" % (", ".join(trg_args)) if len(trg_args) != 0 else ""

    def get_src_fn(self, component):
        fn = component["alex"][0]
        return fn

    @staticmethod
    def get_src_args_str(component):
        src_args = list(component["alex"][1].keys())
        return ", ".join(src_args)

    @staticmethod
    def get_translation(src_fn, src_args, trg_fn, trg_args):
        code_str = "def %s(%s):\n" % (src_fn, src_args)
        code_str += "\treturn %s%s\n\n" % (trg_fn, trg_args)
        return code_str

    def cache_translation(self, block):
        for fn in self.blocks[block]:
            component = self.blocks[block][fn]
            try:
                if "alex" not in component:
                    continue
                src_fn = self.get_src_fn(component)
                src_args_str = self.get_src_args_str(component)
                trg_fn = self.get_trg_fn(component, self.engine)
                trg_args_str = self.get_trg_args_str(component, self.engine)

                code_str = self.get_translation(src_fn,
                                                src_args_str,
                                                trg_fn,
                                                trg_args_str)
                self.alex_defs[block].append(code_str)
                self.inline_index_translation.append(src_fn)

            except Exception as error:
                traceback.print_exc()
                print("%s not implemented" % fn)

    def cache_param_translation(self):
        # Generate transltion for shape function
        for component in const.PARAMS:
            for param in const.PARAMS[component]:
                shape = const.PARAMS[component][param]["shape"]
                src_fn = self.get_src_fn(shape)
                self.inline_index_translation.append(src_fn)

                src_args_str = self.get_src_args_str(shape)
                trg_fn = self.get_trg_fn(shape, self.engine)
                trg_args_str = self.get_trg_args_str(shape, self.engine)

                shape_code_str = self.get_translation(src_fn,
                                                      src_args_str,
                                                      trg_fn,
                                                      trg_args_str)
                self.alex_defs["param"].append(shape_code_str)
                constructor = const.CONSTRUCTORS["params"]
                src_fn = "%s_%s" % (component, param)

                src_args_str = self.get_src_args_str(constructor)
                trg_fn = self.get_trg_fn(constructor, self.engine)
                trg_args_str = self.get_trg_args_str(constructor, self.engine)
                constructor_code_str = self.get_translation(src_fn,
                                                            src_args_str,
                                                            trg_fn,
                                                            trg_args_str)
                self.alex_defs["param"].append(constructor_code_str)
                self.inline_index_translation.append(src_fn)

    def cache_boiler_plate(self):
        imports = NAMESPACES[self.engine].add_imports()
        configs = NAMESPACES[self.engine].add_global_configs()
        return imports + configs

    def generate_alex(self, node):
        try:
            _node = nodes(node).generate_code(node,
                                              self.annotated,
                                              self.engine)
            if _node is None:
                return node
            else:
                node = _node
            if node["code"]["type"] == FUNCTION:
                _param_node = self.annotated[self.annotated[node["parent"]]["parent"]]

                if node["value"] in const.ALL_INITIALIZERS \
                   and self.load \
                   and _param_node["parent"] in self.ckpt.matched:
                    _name = "%s/%s" % (_param_node["parent"],
                                       _param_node["value"])
                    code_str = "%s = %s(inputs=%s, dtype='%s', device=device)\n" % (node["code"]["return_symbol"],
                                                                                  "as_tensor",
                                                                                  "ckpt['%s']" % _name,
                                                                                  node["dtype"])
                else:
                    code_str = "%s = %s\n" % (node["code"]["return_symbol"],
                                              node["code"]["str"])

                block = self.get_block(node)
                if node["code"]["inline"]:
                    self.inline_index_python.append(_name_strs_to_names(node["code"]["return_symbol"]))
                    self.alex_inline[block].append(code_str)
                else:
                    self.alex_code[block].append(code_str)
                    if node["code"]["extra"]:
                        self.alex_code[block].append(node["code"]["extra"])

        except Exception as err:
            traceback.print_exc()
        return node

    def generate_python(self):
        self.annotate_tree()
        boiler_str = self.cache_boiler_plate()

        param_args = ["ckpt=None"] # if self.load else []
        param_str = self.get_dl_code(block="param",
                                     fn_name="get_trainable_params",
                                     return_str="trainable_params",
                                     manual_args=param_args,
                                     prefix="trainable_params = dict()\n")
        loss_str = self.get_dl_code(block="loss",
                                    fn_name="get_loss",
                                    manual_args=["trainable_params", "inputs"])
        optimizer_str = self.get_dl_code(block="optimizer",
                                         fn_name="get_optimizer",
                                         manual_args=["trainable_params", ])
        component_str = self.get_dl_code(block="model",
                                         fn_name="model",
                                         manual_args=["input_data",
                                                      "trainable_params",
                                                      "training"])
        if len(self.blocks["scheduler"]) != 0:
            scheduler_str = self.get_dl_code(block="scheduler",
                                             fn_name="get_scheduler",
                                             manual_args=["optimizer"])
        else:
            scheduler_str = ""

        code = NAMESPACES[self.engine].wrap_in_class(param_str,
                                                     component_str,
                                                     loss_str,
                                                     optimizer_str,
                                                     scheduler_str)
        util.write_to(boiler_str, self.output_file)
        util.write_to(code, self.output_file)

    def get_dl_code(self, block, fn_name, manual_args=[], return_str=None, prefix=None):

        util.concatenate_files([self.cache_def_path],
                               self.filepaths[block])
        if prefix is not None:
            write_list_to_file(prefix, self.filepaths[block])

        write_list_to_file(self.alex_inline[block], self.filepaths[block])
        write_list_to_file(self.alex_code[block], self.filepaths[block])
        filename = self.filepaths[block].split("/")[-1]
        dirname = "/".join(self.filepaths[block].split("/")[:-1])
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            warnings.warn("deprecated", DeprecationWarning)
            inline(self.inline_index_translation, dirname, filename)
            inline(self.inline_index_python, dirname, filename)
        with open(self.filepaths[block], "r") as f:
            src_code = f.readlines()
        # TODO: automatically detect args
        if len(src_code) != 0:
            src_code = ns_alex.wrap_in_function(src_code, fn_name, manual_args, return_str)
        else:
            src_code = ""
        return src_code


# -------------------------- Inline -------------------------------
def write_list_to_file(lst, filename):
    util.write_to("".join(lst), filename)
    # print("Written to %s" % filename)


def inline(inline_fns, dirname, filename):
    project = Project(dirname)
    resource = project.get_resource(filename)
    for inline_fn in inline_fns:
        try: # FIXME: temporarily handling not implemented error
            changes = create_inline(project,
                                    resource,
                                    resource.read().index(inline_fn)).get_changes(remove=True)
            project.do(changes)
        except Exception as e:
            pass
            # print("%s not implemented" % (inline_fn))
            # print("Error: %s", e)



import pkg_resources
import os


COMPONENT_BASE_PATH = pkg_resources.resource_filename("alex", "./components")
CACHE_BASE_PATH = pkg_resources.resource_filename("alex", "../cache")
EXAMPLES_PATH = pkg_resources.resource_filename("alex", "../examples")
CHECKPOINT_BASE_PATH = pkg_resources.resource_filename("alex", "..checkpoints/")
ENGINE_PATH = pkg_resources.resource_filename("alex", "./engine")

ALEX_CACHE_BASE_PATH = ".alex.cache/"
os.makedirs(ALEX_CACHE_BASE_PATH, exist_ok=True)

DEFAULT_DTYPE = "float32"
# ---------------------------- Trainable parameters ------------------------- #
CONSTRUCTORS = {"params": {"tf": ["tf.Variable",
                                  {"initial_value": "initializer",
                                   "trainable": "is_trainable",
                                   "caching_device": "None",
                                   "name": "name",
                                   "variable_def": "None",
                                   "dtype": "tf_dtypes[dtype]",
                                   "import_scope": "None",
                                   "constraint": "None",
                                   "synchronization": "tf.VariableSynchronization.AUTO",
                                   "shape": "None"}],
                           "keras": ["", []],
                           "pytorch": ["torch.nn.parameter.Parameter",
                                       {"data": "initializer",
                                        "requires_grad": "is_trainable"}],
                           "alex": ["params",
                                    {"initializer": False,
                                     "name": True,
                                     "dtype": True,
                                     "is_trainable": True}]}}


# TODO: as it is now, the param names need to be globally unique
PARAMS = {# "": {"linear"},
          "conv": {"filters": {"derivative": True,
                               "ingredient": "conv",
                               "shape": {"pytorch": ["[n_filters, input_shape, kernel_size_h, kernel_size_w]",
                                                     {}],
                                         "tf": ["(kernel_size_h, kernel_size_w, input_shape, n_filters)",
                                                {}],
                                         "alex": ["conv_filters_shape",
                                                  {"n_filters": True,
                                                   "input_shape": False,
                                                   "kernel_size_h": True,
                                                   "kernel_size_w": True}]}}},
          "dense": {"weights": {"derivative": True,
                                "ingredient": "dense",
                                "shape":
                                {"pytorch": ["[n_units, input_shape]",
                                             {}],
                                 "tf": ["[input_shape, n_units]",
                                        {}],
                                 "alex": ["dense_weights_shape",
                                          {"n_units": True,
                                           "input_shape": False}]}},
                    "bias": {"derivative": True,
                             "ingredient": "dense",
                             "shape":
                             {"alex": ["dense_bias_shape",
                                          {"dim": True}],
                              "pytorch": ["[dim, ]",
                                          {}],
                              "tf": ["[dim, ]",
                                     {}]}}},
          "batch_normalize": {"mean": {"derivative": False,
                                       "ingredient": "batch_normalize",
                                       "shape": {"pytorch":
                                                 ["[input_shape, ]",
                                                  {}],
                                                 "tf":
                                                 ["[input_shape, ]",
                                                  {}],
                                                 "alex":
                                                 ["batch_normalize_mean_shape",
                                                  {"input_shape": False}]}},
                              "variance": {"derivative": False,
                                           "ingredient": "batch_normalize",
                                           "shape": {"pytorch":
                                                     ["[input_shape, ]",
                                                      {}],
                                                     "tf":
                                                     ["[input_shape, ]",
                                                      {}],
                                                     "alex":
                                                     ["batch_normalize_variance_shape",
                                                      {"input_shape": False}]}},
                              "offset": {"derivative": True,
                                         "ingredient": "batch_normalize",
                                         "shape": {"pytorch":
                                                   ["[input_shape, ]",
                                                    {}],
                                                   "tf":
                                                   ["[input_shape, ]",
                                                    {}],
                                                   "alex":
                                                   ["batch_normalize_offset_shape",
                                                    {"input_shape": False}]}},
                              "scale": {"derivative": True,
                                        "ingredient": "batch_normalize",
                                        "shape": {"pytorch":
                                                  ["[input_shape, ]",
                                                   {}],
                                                  "tf":
                                                  ["[input_shape, ]",
                                                   {}],
                                                  "alex":
                                                  ["batch_normalize_scale_shape",
                                                   {"input_shape": False}]}},
          }
}

ALL_PARAMS = dict()
ALL_TRAINABLE_PARAMS = dict()
ALL_OTHER_PARAMS = dict()
for ingredient in PARAMS:
    for param in PARAMS[ingredient]:
        ALL_PARAMS[param] = PARAMS[ingredient][param]
        if PARAMS[ingredient][param]["derivative"]:
            ALL_TRAINABLE_PARAMS[param] = PARAMS[ingredient][param]
        else:
            ALL_OTHER_PARAMS[param] = PARAMS[ingredient][param]
ALL_PARAMS_LIST = list(ALL_PARAMS.keys())
ALL_TRAINABLE_PARAMS_LIST = list(ALL_TRAINABLE_PARAMS.keys())
ALL_OTHER_PARAMS_LIST = list(ALL_OTHER_PARAMS.keys())

INITIALIZERS_NO_HYPE = {"zeros_initializer": {"keras": ["",
                                                        []],
                                              "tf": ["tf.zeros_initializer()",
                                                     {"shape": "shape"}],
                                              "pytorch": ["torch.nn.init.zeros_",
                                                          {"tensor": "torch.empty(*shape)"}],
                                              "alex": ["zeros_initializer",
                                                       {"dtype": True,
                                                        "shape": False}]},
                        "ones_initializer": {"keras": ["",
                                                       []],
                                             "tf": ["tf.ones_initializer()",
                                                    {"shape": "shape"}],
                                             "pytorch": ["torch.nn.init.ones_",
                                                         {"tensor": "torch.empty(*shape)"}],
                                             "alex": ["ones_initializer",
                                                      {"dtype": True,
                                                       "shape": False}]}}

INITIALIZERS = {"he_uniform": {"keras": ["keras.initializers.he_uniform", []],
                               "tf": ["tf.keras.initializers.he_uniform(seed=seed)",
                                      {"shape": "shape"}],
                               "pytorch": ["torch.nn.init.kaiming_uniform",
                                           {"tensor": "torch.empty(*shape)"}],
                               "alex": ["he_uniform",
                                        {"dtype": True,
                                         "seed": True,
                                         "shape": False}]},
                "xavier_uniform": {"keras": ["keras.initializers.glorot_uniform",
                                             {}],
                                   "tf": ["tf.keras.initializers.glorot_uniform(seed=seed)",
                                          {"shape": "shape"}],
                                   "pytorch": ["torch.nn.init.xavier_uniform_",
                                               {"tensor": "torch.empty(*shape)"}],
                                   "alex": ["xavier_uniform",
                                            {"dtype": True,
                                             "seed": True,
                                             "shape": False}]}}

ALL_INITIALIZERS = {**INITIALIZERS_NO_HYPE, **INITIALIZERS}

COMPONENTS = {"conv": {"tf": ["tf.nn.conv2d",
                              {"input": "inputs",
                               "filters": "filters",
                               "strides": "strides",
                               "padding": "padding",
                               "data_format": "'NHWC'",
                               "dilations": "dilation",
                               "name": "name"}],
                       "keras": ["keras.layers.Conv2D", {}],
                       "pytorch": ["torch.nn.functional.conv2d",
                                   {"input": "inputs",
                                    "weight": "filters",
                                    "bias": "None",
                                    "stride": "strides",
                                    "padding": "padding",
                                    "dilation": "dilation",
                                    "groups": "1"}],
                       "alex": ["conv",
                                {"filters": False,
                                 "padding": True,
                                 "strides": True,
                                 "dilation": True,
                                 "inputs": False,
                                 "dtype": True,
                                 "name": True,
                                 "training": False}]},
              "dense": {"tf": ["tf.add(x=tf.matmul(a=inputs, b=weights), y=bias, name=name)",
                               {}], # FIXME
                        "keras": ["keras.layers.Dense", []],
                        "pytorch": ["torch.nn.functional.linear",
                                    {"weight": "weights",
                                     "bias": "bias",
                                     "input": "inputs"}],
                        "alex": ["dense",
                                 {"inputs": False,
                                  "weights": False,
                                  "bias": False,
                                  "dtype": True,
                                  "name": True}]},
              "batch_normalize": {"tf": ["tf.nn.batch_normalization",
                                         {"x": "inputs",
                                          "mean": "mean",
                                          "variance": "variance",
                                          "offset": "offset",
                                          "scale": "scale",
                                          "variance_epsilon": "epsilon",
                                          "name": "name"}],
                                  "pytorch": ["torch.nn.functional.batch_norm",
                                              {"input": "inputs",
                                               "running_mean": "mean",
                                               "running_var": "variance",
                                               "weight": "scale",
                                               "bias": "offset",
                                               "training": "training",
                                               "momentum": "momentum", "eps": "epsilon"}],
                                  "alex": ["batch_normalize",
                                           {"mean": False,
                                            "variance": False,
                                            "offset": False,
                                            "scale": False,
                                            "momentum": True,
                                            "epsilon": True,
                                            "inputs": False,
                                            "name": True,
                                            "dtype": True,
                                            "training": False}]}
}

# 'deconv': {"tf": "tf.nn.conv2d_transpose"},

# ---------------------------- Deterministic layers -------------------------- #
DET_COMPONENTS_HYPE = {"dropout": {"tf": ["tf.nn.dropout",
                                          {"x": "inputs",
                                           "rate": "dropout_rate",
                                           "noise_shape": "None",
                                           "seed": "None",
                                           "name": "name"}],
                                   "keras": ["keras.layers.Dropout", []],
                                   "pytorch": ["torch.nn.functional.dropout",
                                               {"input": "inputs",
                                                "p": "dropout_rate",
                                                "training": "training",
                                                "inplace": "False"}],
                                   "alex": ["dropout",
                                            {"dropout_rate": True,
                                             "inputs": False,
                                             "name": True,
                                             "dtype": True,
                                             "training": False}]},
                       "max_pool2d": {"tf": ["tf.nn.max_pool",
                                             {"input": "inputs",
                                              "ksize": "window_shape",
                                              "strides": "strides",
                                              "padding": "padding",
                                              "data_format": "'NHWC'",
                                              "name": "name"}],
                                      "pytorch": ["torch.nn.functional.max_pool2d",
                                                  {"input": "inputs",
                                                   "kernel_size": "window_shape",
                                                   "stride": "strides",
                                                   "padding": "padding"}],
                                      "alex": ["max_pool2d",
                                               {"window_shape": True,
                                                "strides": True,
                                                "padding": True,
                                                "dilation": True,
                                                "inputs": False,
                                                "dtype": True,
                                                "name": True}]}}

DET_COMPONENTS_NO_HYPE = {"relu": {"tf": ["tf.nn.relu",
                                          {"name": "name",
                                           "features": "inputs"}],
                                   "keras": ["keras.layers.ReLU", []],
                                   "pytorch": ["torch.nn.functional.relu",
                                               {"input": "inputs",
                                                "inplace": "False"}],
                                   "alex": ["relu",
                                            {"inputs": False,
                                             "dtype": True,
                                             "name": True}]},
                          "softmax": {"tf": ["tf.nn.softmax",
                                             {"logits": "inputs",
                                              "name": "name"}],
                                      "pytorch": ["torch.nn.functional.softmax",
                                                  {"input": "inputs",
                                                   "dim": "None"}],
                                      "alex": ["softmax",
                                               {"inputs": False,
                                                "dtype": True,
                                                "name": True}]},
                          "sigmoid": {"tf": ["tf.math.sigmoid",
                                             {"x": "inputs",
                                              "name": "name"}],
                                      "pytorch": ["torch.sigmoid",
                                                  {"input": "inputs"}],
                                      "alex": ["sigmoid",
                                               {"inputs": False,
                                                "dtype": True,
                                                "name": True}]},
                          "flatten": {"tf": ["tf.reshape",
                                             {"tensor": "inputs",
                                              "shape": "(-1, tf.math.reduce_prod(tf.convert_to_tensor(shape)))",
                                              "name": "name"}],
                                      "keras": ["keras.layers.Flatten", []],
                                      "pytorch": ["torch.flatten",
                                                  {"input": "inputs",
                                                   "start_dim": "1",
                                                   "end_dim": "-1"}],
                                      "alex": ["flatten",
                                               {"inputs": False,
                                                "shape": False,
                                                "dtype": True,
                                                "name": True}]},
                          "add": {"tf": ["tf.math.add",
                                         {"x": "inputs[0]",
                                          "y": "inputs[1]",
                                          "name": "name"}],
                                  "pytorch": ["torch.add",
                                              {"input": "inputs[0]",
                                               "other": "inputs[1]"}],
                                  "alex": ["add",
                                           {"inputs": False,
                                            "dtype": True,
                                            "name": True}]},
                          "zeros": {"tf": ["tf.zeros",
                                           {}],
                                    "pytorch": ["torch.zeros",
                                                {"input_shape": "shape"}],
                                    "alex": ["zeros",
                                             {"shape": False,
                                              "name": True}]}}
# TODO: concatenate, softmax
DL_LAYERS = {**COMPONENTS,
             **DET_COMPONENTS_HYPE,
             **DET_COMPONENTS_NO_HYPE}

INFERENCE = {} # inference function
LOSSES = {"cross_entropy": {"tf": ["tf.nn.softmax_cross_entropy_with_logits",
                                   {"labels": "inputs[0]",
                                    "logits": "inputs[1]",
                                    "axis": "-1",
                                    "name": "name"}],
                            "pytorch": ["torch.nn.functional.cross_entropy",
                                        {"weight": "None",
                                         "ignore_index": "-100",
                                         "reduction": "'mean'",
                                         "target": "inputs[0]",
                                         "input": "inputs[1]"}],
                            "alex": ["cross_entropy",
                                     {"name": True,
                                      "dtype": True,
                                      "training": False}]}}
# TODO: bitwise, l2, mean_square
REGULARIZERS = {"regularizer_l2": {"tf": ["coeff*sum(list(map(lambda x: tf.nn.l2_loss(t=trainable_params[x], name=name), inputs)))",
                                          {}],
                                   "pytorch": ["coeff*sum(list(map(lambda x: torch.norm(input=trainable_params[x]), inputs)))",
                                               {}],
                                   "keras": ["tf.keras.regularizers.L2", []],
                                   "alex": ["regularizer_l2",
                                            {"coeff": True,
                                             "name": True,
                                             "training": False,
                                             "dtype": True,
                                             "trainable_params": False,
                                             "inputs": False}]}}

OPTIMIZERS = {"adam": {"tf": ["tf.optimizers.Adam",
                              {"learning_rate": "learning_rate",
                               "beta_1": "beta1",
                               "beta_2": "beta2",
                               "epsilon": "epsilon",
                               "name": "name"}],
                       "pytorch": ["torch.optim.Adam",
                                   {"params": "trainable_params",
                                    "lr": "learning_rate",
                                    "betas": "(beta1, beta2)",
                                    "eps": "epsilon"
                                   }],
                       "alex": ["adam",
                                {"learning_rate": True,
                                 "beta1": True,
                                 "beta2": True,
                                 "epsilon": True,
                                 "inputs": False,
                                 "name": True,
                                 "training": False,
                                 "dtype": True,
                                 "trainable_params": False,
                                 "initial_learning_rate": True}]}
}
# TODO: gradient_descent, momentum_optimizer
LOSS_BLOCK = {**LOSSES, **REGULARIZERS}
INGREDIENT_TYPES = {**DL_LAYERS, **LOSSES, **OPTIMIZERS, **INFERENCE}

INPUT_TYPES = ("data", "label")


LEARNING_RATE_DECAY = {"exponential_decay": {"tf": ["tf.keras.optimizers.schedules.ExponentialDecay",
                                                    {"initial_learning_rate": "initial_learning_rate",
                                                     "decay_steps": "decay_steps",
                                                     "decay_rate": "decay_rate",
                                                     "staircase": "staircase"}],
                                             "pytorch": ["torch.optim.lr_scheduler.ExponentialLR",
                                                         {"optimizer": "optimizer",
                                                          "gamma": "decay_rate",
                                                          "last_epoch": "-1",
                                                          "verbose": "False"
                                                         }],
                                             "alex": ["exponential_decay",
                                                      {"initial_learning_rate": True,
                                                       "decay_steps": True,
                                                       "decay_rate": True,
                                                       "staircase": True,
                                                       "optimizer": False
                                                      }]}} # FIXME
SCHEDULER_BLOCK = {**LEARNING_RATE_DECAY}

OPTIMIZER_BLOCK = {**OPTIMIZERS}

TENSOR_SHAPE = {"input_shape": {"tf": ["inputs.get_shape().as_list()", []]}}

AS_TENSOR = {"as_tensor": {"tf": ["tf.convert_to_tensor",
                                  {"value": "np.asarray(inputs)",
                                   "dtype": "tf_dtypes['dtype']",
                                   "dtype_hint": "None"}],
                           "pytorch": ["torch.as_tensor",
                                       {"data": "np.asarray(inputs)",
                                        "dtype": "torch_types[dtype]",
                                        "device": "device"}],
                           "alex": ["as_tensor", {"inputs": False,
                                                  "dtype": True,
                                                  "device": False,
                                                  "name": True}]}}
MODEL_BLOCK = {**DL_LAYERS, **AS_TENSOR}

UTILITY = {**TENSOR_SHAPE, **LEARNING_RATE_DECAY, **AS_TENSOR}

ALL_COMPONENTS = {**COMPONENTS,
                  **ALL_INITIALIZERS,
                  **DET_COMPONENTS_HYPE,
                  **DET_COMPONENTS_NO_HYPE,
                  **INFERENCE,
                  **LOSSES,
                  **REGULARIZERS,
                  **OPTIMIZERS,}
                  # **UTILITY}
COMPONENT_RECIPES = {"adense",
                     "resnet_16",
                     "resnet_32", "resnet_32_short_cut",
                     "resnet_64", "resnet_64_short_cut",
                     "resnet_128", "resnet_128_short_cut",
                     "resnet_256", "resnet_256_short_cut",
                     "resnet_512", "resnet_512_short_cut"}

ALL_RECIPES = {"root",
               *COMPONENT_RECIPES}

TYPES = {0: "DETERMINISTIC_COMPONENTS_WITHOUT_HYPERPARAMETER",
         1: "DETERMINISTIC_COMPONENTS_WITH_HYPERPARAMETER",
         2: "TRAINABLE_COMPONENTS"}

# ------------------------- User defined functions --------------------------- #

USER_FNS = {"lr_scheduler": "engine.ns_user.learning_rate"}


FNS = [*list(ALL_COMPONENTS.keys()),
       *list(USER_FNS.keys()),
       *list(SCHEDULER_BLOCK.keys()),
       *ALL_TRAINABLE_PARAMS, "shape"]

# -------------------------------- String consts ----------------------------- #
# Network build related
COMPONENT_LIST = 'component_list'
LOSS_LIST = 'loss_list'

YAML = ".yml"
COMPONENT_NAMES = 'component_types'
NAME = 'name'
DEFINITION = 'definition'
REPEAT = 'repeat'
SHARE = "share"
KEEP_TRACK = 'keep_track'
RUNTIME = "runtime"
RUNTIMES = "runtimes"
DATA = "data"
LABEL = "label"
# General config
LOGS = 'logs'
VAR = 'var'
VAR_NAME = "var_name"
STATS = "stats"
COMPONENT = 'component'
LOSS = 'loss'
SOLVER = 'optimizers'
TYPE = 'type'
INPUTS = 'inputs'
PARAMS_NETWORK = 'params_network'
PARAMS_TRAINING = "params_training"
PARAMS_LOSSES = 'params_losses'
PARAMS_SOLVERS = 'params_optimizers'
HYPERPARAMS = 'hyperparams'
IS_TRAINING = 'is_training'
LEARNING_RATE = 'learning_rate'
CONFIG = "config"
STAT = "stat"
VALUE = "value"
TENSOR = "tensor"
CHECKPOINT_DIR = "checkpoint_dir"
CHECKPOINT_EXT = ".json"
INDEX = "index"
EVAL = "eval"
INFERENCE = 'inference'
###############################################
# Graph related:
META = "meta"
SUBGRAPH = "subgraph"
GRAPH = "graph"
ROOT = "root"
SCOPE = "scope"
LEVEL = "level"
SEQUENTIAL = "sequential"
###############################################
# Component related
# Component hyperparams
### general
ACTIVATION_FUNCTION = 'activation_function'
INITIALIZATION = 'initialization'
INITIALIZER = "initializer"
RANDOM = 'random'
OUTPUT_SHAPE = 'output_shape'
SHAPE = "shape"
### conv2d
STRIDES = 'strides'
FILTERS = 'filters'
N_FILTERS = 'n_filters'
KERNEL_SIZE = 'kernel_size'

### dense
N_UNITS = 'n_units'

### batch_norm
VARIANCE_EPSILON = "variance_epsilon"
DECAY = "decay"
# pool:
WINDOW_SHAPE = 'window_shape'
POOLING_TYPE = 'pooling_type'
PADDING = 'padding'

### training
WEIGHT_DECAY = "weight_decay"
DROPOUT_RATE = 'dropout_rate'

###############################################
# Data related
DATA_NAME = 'data_name'
TRAINING_DATA = 'training_data'
TESTING_DATA = 'testing_data'
INPUT_DATA = 'input_data'
LABELS = 'labels'
MEAN = 'mean'
STD = 'std'
DTYPE = "dtype"
# Data hyperparams
BATCH_SIZE = 'batch_size'
TEST_SIZE = 'test_size'

################################################
## 20200709
RPT_IDX = "__rpt_idx__" # repeat index

COMPONENT_INFO_KEYS = [VALUE, HYPERPARAMS, STAT, NAME, VAR]

#################################################
# alex namespace
# arguments
ALEX_ARG_NAME = "name"
ALEX_ARG_INPUTS = "inputs"
ALEX_ARG_LABELS = "labels"
ALEX_ARG_INITIAL_VALUE = "initial_value"
ALEX_ARG_INPUT_SHAPE = "input_shape"
ALEX_ARG_SHAPE = "shape"
ALEX_ARG_TRAINABLE_PARAMS = "trainable_params"
ALEX_ARG_TRAINABLE_PARAMS_NAME = "trainable_params_name"



# ----------------------------------------------------------------------
# Created: sön apr 25 14:43:00 2021 (+0200)
# Last-Updated:
# Filename: schema.py
# Author: Yinan Yu
# Description:
# ----------------------------------------------------------------------

from glob import glob
from pprint import pprint
import os
from copy import deepcopy

from alex.alex import const, util

from jsonschema import validate


if __name__ == "__main__":

    schema = util.read_yaml(util.get_component_spec_path("conv"))["definitions"]["conv"]
    conv2d = util.read_yaml(util.get_component_path("conv"))
    conv2d["invalid_key"] = None
    validate(instance=conv2d, schema=schema)



# ----------------------------------------------------------------------
# Created: mån jul 12 00:52:56 2021 (+0200)
# Last-Updated:
# Filename: validation.py
# Author: Yinan Yu
# Description:
# ----------------------------------------------------------------------
import numpy as np
from math import floor, ceil
from pprint import pprint
import traceback
from copy import deepcopy
from jsonschema import validate
from alex.alex import const, util
import os


class Ingredient():
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.transformation = None

    def load_schema(self, component_type):
        schema_path = os.path.join(const.COMPONENT_BASE_PATH,
                                   "." + component_type + ".yml")
        try:
            schema = util.read_yaml(schema_path)["definitions"][component_type]
        except Exception:
            schema = None
            # TODO:
            # print("%s schema not implemented" % component_type)
        return schema

    def eval_input_channels(self, component, components):
        inputs = self.get_input_shape(component, components)
        if inputs is not None:
            util.replace_key(component["value"]["hyperparams"],
                             "input_shape",
                             inputs[0][-1])

    def get_input_shape(self, component, components):
        if component["meta"]["inputs"] is None or component["meta"]["inputs"] == "inputs":
            return None

        return list(map(lambda x: components[x]["meta"]["shape"] if x in components else None,
                        component["meta"]["inputs"]))

    def get_shape(self, component, components):
        return self.get_input_shape(component, components)[0]

    def get_transformation(self):
        return self.transformation

    def get_dimensions(self, component, components):
        input_shape = self.get_input_shape(component, components)
        return [input_shape, # input shape
                self.get_shape(component, components)] # output shape

    def validate_schema(self, component, schema):
        validate(instance=component["value"]["hyperparams"], schema=schema)

    def validate_connection(self, component, components):
        transformations = self.get_transformation()
        dims = self.get_dimensions(component, components)
        if transformations is None:
            return True
        else:
            valid = False
            msg = ""
            for trans in transformations:
                if valid:
                    return valid
                try:
                    if len(trans[0]) != len(dims[0]):
                        msg = "Number of inputs mismatch %s (ingredient: %s, dim %s)" % (component["meta"]["name"],
                                                                                         component["meta"]["type"],
                                                                         str(dims))
                        raise Exception(msg)
                    elif len(list(filter(lambda x: len(x[0]) != len(x[1]),
                                     zip(trans[0], dims[0]))))!=0:
                        msg = "Tensor rank mismatch %s (ingredient: %s, dim: %s)" % (component["meta"]["name"],
                                                                                    component["meta"]["type"],
                                                                                     str(dims))
                        raise Exception(msg)
                    else:
                        trans_ = [i for sub in trans[0] for i in sub] + trans[1]
                        dims_ = [i for sub in dims[0] for i in sub] + dims[1]
                        n = len(trans_)
                        pattern_required = [trans_[i]==trans_[i+j] for i in range(n-1) for j in range(1, n-i)]
                        pattern_data = [dims_[i]==dims_[i+j] for i in range(n-1) for j in range(1, n-i)]
                        for r in pattern_required:
                            if pattern_required and (not pattern_data):
                                msg = "Tensor dimension mismatch" % component["meta"]["name"]
                                raise Exception(msg)
                    valid = True
                except:
                    pass
                    # traceback.print_exc()
                    # raise Exception(msg)
            if not valid:
                print(msg)
                raise Exception(msg)


# ------------------------------------------------------------------------ #
class Dense(Ingredient):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.transformation = [[[["a"]], ["b"]]]

    def get_shape(self, component, *args):
        return [int(component["value"]["hyperparams"]["weights"]["shape"]["n_units"]), ]

class CrossEntropy(Ingredient):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.transformation = [[[["a"], ["a"]], ["c"]]]

    def get_shape(self, component, components):
        return [0, ]


class Regularizer(Ingredient):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_shape(self, component, components):
        return [0, ]

    def get_input_shape(self, component, components):
        # if component["meta"]["inputs"] is None:
        #     return None
        # return list(map(lambda x: components[x]["meta"]["shape"],
        #                 component["meta"]["inputs"]))
        return None

class Batchnorm(Ingredient):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.transformation = [[[["a", "b", "c"]], ["a", "b", "c"]]]


class Add(Ingredient):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.transformation = [[[["a", "b", "c"], ["a", "b", "c"]], ["a", "b", "c"]],
                               [[["a", "b"], ["a", "b"]], ["a", "b"]],
                               [[["a"], ["a"]], ["a"]]]

    def get_shape(self, component, components):
        input_shape = self.get_input_shape(component, components)
        if input_shape is None or input_shape[0] is None:
            return None
        return input_shape[0]


class Conv2D(Ingredient):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.transformation = [[[["a", "b", "c"]], ["d", "e", "f"]]]

    def get_shape(self, component, components):
        input_shape = self.get_input_shape(component, components)
        if input_shape is None or input_shape[0] is None:
            return None
        input_shape = input_shape[0]
        hyperparams = component["value"]["hyperparams"]
        strides = hyperparams["strides"]
        dilation = hyperparams["dilation"]
        padding = hyperparams["padding"]
        filters_shape = hyperparams["filters"]["shape"]
        k_h = int(filters_shape["kernel_size_h"])
        k_w = int(filters_shape["kernel_size_w"])
        n_filters = int(filters_shape["n_filters"])
        h = input_shape[0]
        w = input_shape[1]
        if padding == "SAME":
            hout = ceil(h/strides[0])
            wout = ceil(w/strides[1])
            shape = [hout, wout, n_filters]
        elif padding == "VALID":
            padding = [0, 0]
            shape = [int(floor((h + 2*padding[0] - dilation*(k_h-1) - 1)/strides[0] + 1)),
                     int(floor((w + 2*padding[1] - dilation*(k_w-1) - 1)/strides[1] + 1)),
                     int(n_filters)]
        return shape


class MaxPool2D(Ingredient):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.required_input_rank = 3
        self.transformation = [[[["a", "b", "c"]], ["d", "e", "f"]]]

    def get_shape(self, component, components):
        input_shape = self.get_input_shape(component, components)
        if input_shape is None or input_shape[0] is None:
            return None
        input_shape = input_shape[0]
        hyperparams = component["value"]["hyperparams"]
        padding = hyperparams["padding"]
        strides = hyperparams["strides"]
        dilation = hyperparams["dilation"]
        kernel_shape = hyperparams["window_shape"]

        k_h = int(kernel_shape[0])
        k_w = int(kernel_shape[1])
        h = input_shape[0]
        w = input_shape[1]
        if padding == "SAME":
            hout = ceil(h/strides[0])
            wout = ceil(w/strides[1])
            shape = [hout, wout, input_shape[-1]]
        elif padding == "VALID":
            padding = [0, 0]
            shape = [floor((h + 2*padding[0] - dilation*(k_h-1) - 1)/strides[0] + 1),
                     floor((w + 2*padding[1] - dilation*(k_w-1) - 1)/strides[1] + 1),
                     input_shape[-1]]
        return shape


# None trainable
class Flatten(Ingredient):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.required_input_rank = None
        self.transformation = [[[["a", "b", "c"]], ["d"]],
                               [[["a", "b"]], ["d"]]]

    def get_shape(self, component, components):
        shape = list(map(lambda x: components[x]["meta"]["shape"],
                         component["meta"]["inputs"]))
        return [int(np.prod(np.asarray(shape))), ]


class Data(Ingredient):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_shape(self, component, *args):
        shape = component["value"]["hyperparams"]["shape"]
        return [shape["height"], shape["width"], shape["channels"]]


class Label(Ingredient):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_shape(self, component, *args):
        shape = component["value"]["hyperparams"]["shape"]
        return [shape["dim"], ]


INGREDIENTS = {"ingredient": Ingredient(),
               "data": Data(),
               "cross_entropy": CrossEntropy(),
               "label": Label(),
               "dense": Dense(),
               "add": Add(),
               "batch_normalize": Batchnorm(),
               "regularizer_l2": Regularizer(),
               "conv": Conv2D(),
               "max_pool2d": MaxPool2D(),
               "flatten": Flatten()}


def get(component):
    value = component["meta"]["type"]
    # TODO: group ingredient by block
    if value not in INGREDIENTS:
        value = "ingredient"
    return INGREDIENTS[value]



# ----------------------------------------------------------------------
# Created: mån maj 10 01:52:20 2021 (+0200)
# Last-Updated:
# Filename: test_compare.py
# Author: Yinan Yu
# Description:
# ----------------------------------------------------------------------
import unittest
import os
from pprint import pprint
import collections
from copy import deepcopy
from alex.alex import core, compare, const, util


class TestCompare(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.conv = util.read_yaml(util.get_component_path("conv"))
        self.modified_conv = deepcopy(self.conv)
        self.modified_conv["padding"] = "WRONG"
        self.modified_conv["filters"]["nfilters"] = dict()
        self.modified_conv["filters"]["nfilters"]["terrible"] = 5
        self.modified_conv["filters"]["shape"]["added"] = None
        self.modified_conv["filters"]["initializer"]["xavier_uniform"] = "changed"
        self.toy_json = collections.OrderedDict()
        a_subtree = collections.OrderedDict()
        a_subtree["c"] = "l"
        a_subtree["h"] = None
        self.toy_json["f"] = collections.OrderedDict()
        self.toy_json["f"]["a"] = a_subtree
        self.toy_json["f"]["e"] = None
        self.toy_json2 = collections.OrderedDict()
        a_subtree2 = collections.OrderedDict()
        a_subtree2["d"] = "i"
        a_subtree2["c"] = collections.OrderedDict()
        a_subtree2["c"]["j"] = None
        a_subtree2["c"]["b"] = collections.OrderedDict()
        a_subtree2["c"]["b"]["m"] = 5
        self.toy_json2["f"] = collections.OrderedDict()
        self.toy_json2["f"]["a"] = a_subtree2
        self.toy_json2["f"]["k"] = None

    def setUp(self):
        pass

    def test_get_children(self):
        conv_tree = core.json_to_tree(self.conv, "conv")
        conv_key = list(conv_tree.keys())[-1]
        expected_children = ["dilation",
                             "filters",
                             "padding",
                             "strides"]
        children = core.get_children(conv_key, conv_tree, "label")
        self.assertListEqual(children, expected_children)

    def test_json_to_tree(self):
        conv_tree = core.json_to_tree(self.conv, "conv")
        toy_tree = core.json_to_tree(self.toy_json)

    def test_get_keyroots(self):
        toy_tree = core.json_to_tree(self.toy_json)
        keyroots = compare.get_keyroots(toy_tree)
        self.assertListEqual(keyroots, [2, 4, 5])

    def test_ted(self):
        toy_tree = core.json_to_tree(self.toy_json)
        toy_tree2 = core.json_to_tree(self.toy_json2)
        costs, operations = compare.ted(toy_tree, toy_tree2)
        annotation = compare.annotate_ops(operations)
        self.assertEqual(costs, 8)
        core.draw(toy_tree, os.path.join(const.CACHE_BASE_PATH,
                                         "toytree1.png"),
                  annotation[1])
        core.draw(toy_tree2, os.path.join(const.CACHE_BASE_PATH,
                                          "toytree2.png"),
                  annotation[2])

    def test_ted_on_conv(self):
        conv_tree1 = core.json_to_tree(self.conv, "conv")
        conv_tree2 = core.json_to_tree(self.modified_conv, "conv")
        costs, operations = compare.ted(conv_tree1, conv_tree2)
        self.assertEqual(costs, 7)
        annotation = compare.annotate_ops(operations)
        core.draw(conv_tree1,
                  os.path.join(const.CACHE_BASE_PATH,
                               "conv1.png"),
                  annotation[1])
        core.draw(conv_tree2,
                  os.path.join(const.CACHE_BASE_PATH,
                               "conv2.png"),
                  annotation[2])

    def test_dist(self):
        matched = compare.matched_ingredients("examples/configs/small1.yml",
                                              "examples/configs/small2.yml",
                                              render_to="./cache/subtree.png")
        print("Matched ingredients")
        pprint(matched)


if __name__ == '__main__':
    unittest.main()
